! Module to compute electrostatic Coulomb potential
module MW_coulomb_keq0_pim
   use MW_kinds, only: wp
   use MW_ewald, only: MW_ewald_t
   use MW_box, only: MW_box_t
   use MW_constants, only: pi
   use MW_electrode, only: MW_electrode_t
   use MW_ion, only: MW_ion_t
   use MW_localwork, only: MW_localwork_t
   implicit none
   private

   ! Public subroutines
   ! ------------------
   public :: gradQelec_electricfield !grad_Q [grad_mu [U^k=0_muQ]]
   public :: gradmumelt_electricfield !grad_mu [grad_mu [U^k=0_mumu]]
   public :: mumelt2Qelec_potential !grad_Q [U^k=0_muQ]
   public :: qmelt2mumelt_electricfield !-grad_mu [U^k=0_qmu]
   public :: mumelt2mumelt_electricfield !-grad_mu [U^k=0_mumu]
   public :: Qelec2mumelt_electricfield !-grad_mu [U^k=0_muQ]
   public :: melt_forces !-grad_r [(U^k=0_qq + U^k=0_qmu + U^k=0_mumu) + (U^k=0_QQ) + (U^k=0_qQ + U^k=0_muQ)]
   public :: energy !(U^k=0_qq + U^k=0_qmu + U^k=0_mumu) + (U^k=0_QQ) + (U^k=0_qQ + U^k=0_muQ)

contains

   subroutine gradQelec_electricfield(localwork, ewald, box, ions, xyz_ions, &
      electrodes, xyz_atoms, gradQ_efield)
      implicit none
      ! Parameters in
      ! -------------
      type(MW_localwork_t), intent(in) :: localwork       !< Local work assignment
      type(MW_ewald_t),     intent(in) :: ewald          !< Ewald summation parameters
      type(MW_box_t),       intent(in) :: box            !< Simulation box parameters
      type(MW_ion_t),       intent(in) :: ions(:)        !< ions parameters
      real(wp),             intent(in) :: xyz_ions(:,:)  !< ions xyz positions
      type(MW_electrode_t), intent(in) :: electrodes(:)   !< electrode parameters
      real(wp),             intent(in) :: xyz_atoms(:,:)  !< electrode atoms xyz positions

      ! Parameters out
      ! --------------
      real(wp), intent(inout) :: gradQ_efield(:,:) !< Electric field

      ! Local
      ! -----
      integer :: num_ions, num_ion_types, num_elec_types
      integer :: i, j, jz, itype, jtype
      real(wp) :: volfactor
      real(wp) :: zij, erf_ij
      real(wp) :: alpha
      integer :: num_block_full
      integer :: iblock, iblock_offset, istart, iend, jstart, jend
      integer :: count_n, offset_n, count_m, offset_m

      num_ions = size(gradQ_efield,1)/3
      num_ion_types = size(ions,1)
      num_elec_types = size(electrodes,1)

      ! Precompute some factors
      ! -----------------------
      alpha = ewald%alpha

      ! Long range contribution from k=0 mode
      ! -------------------------------------
      ! Normalization
      volfactor = 2.0_wp/box%area(3)


      do itype = 1, num_elec_types
         count_n = electrodes(itype)%count
         offset_n = electrodes(itype)%offset
         do jtype = 1, num_ion_types
            count_m = ions(jtype)%count
            offset_m = ions(jtype)%offset
            ! Number of blocks with full atom2ion interactions
            num_block_full = localwork%pair_ion2atom_num_block_full_local(itype, jtype)
            iblock_offset = localwork%pair_ion2atom_full_iblock_offset(itype, jtype)
            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
               offset_n, count_n, offset_m, count_m, &
               istart, iend, jstart, jend)
               do i = istart, iend
                  do j = jstart, jend
                     jz = j + num_ions*2
                     zij = xyz_atoms(i,3) - xyz_ions(j,3)
                     erf_ij = erf(alpha*zij)
                     gradQ_efield(jz,i) = gradQ_efield(jz,i) + volfactor*pi*erf_ij
                  end do
               end do
            end do
         end do
      end do

   end subroutine gradQelec_electricfield

   !================================================================================
   ! Compute the electric fields felt by each melt ions due to other melt ions
   ! dipoles
   subroutine gradmumelt_electricfield(localwork, ewald, box, ions, xyz_ions, &
      gradmu_efield)
      implicit none

      ! Parameters Input
      ! -------------
      type(MW_localwork_t), intent(in) :: localwork      !< Local work assignment
      type(MW_ewald_t),     intent(in) :: ewald          !< Ewald summation parameters
      type(MW_box_t),       intent(in) :: box            !< Simulation box parameters
      type(MW_ion_t),       intent(in) :: ions(:)        !< ions parameters
      real(wp),             intent(in) :: xyz_ions(:,:)  !< ions xyz positions

      ! Parameters Input/Output
      ! --------------
      real(wp), intent(inout) :: gradmu_efield(:,:)      !< matrix containing the
      !< gradient of the field
      ! Local
      ! -----
      integer :: num_ion_types, num_ions
      integer :: i, j, itype, jtype, iz, jz
      real(wp) :: volfactor
      real(wp) :: zij, zijsq, exp_ij
      real(wp) :: alpha, alphasq, alphapi, sqrpialpha,sqrpi
      integer :: num_block_diag, num_block_full
      integer :: iblock, iblock_offset, istart, iend, jstart, jend
      integer :: count_n, offset_n, count_m, offset_m

      num_ion_types = size(ions,1)
      num_ions = size(gradmu_efield,1)/3

      ! Precompute some factors
      ! -----------------------
      alpha = ewald%alpha
      alphasq = alpha*alpha
      sqrpi=sqrt(pi)
      alphapi = 2.0_wp * alpha / sqrpi
      sqrpialpha = sqrpi / alpha

      ! Long range contribution from k=0 mode
      ! -------------------------------------
      ! Normalization
      volfactor = 2.0_wp/box%area(3)

      do itype = 1, num_ion_types
         count_n = ions(itype)%count
         offset_n = ions(itype)%offset
         do jtype = 1, itype - 1
            count_m = ions(jtype)%count
            offset_m = ions(jtype)%offset
            ! Number of blocks with full ion2ion interactions
            num_block_full = localwork%pair_ion2ion_num_block_full_local(itype, jtype)
            iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype, jtype)

            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
               offset_n, count_n, offset_m, count_m, &
               istart, iend, jstart, jend)
               do i = istart, iend
                  iz = i + num_ions*2
                  do j = jstart, jend
                     jz = j + num_ions*2
                     zij = xyz_ions(j,3) - xyz_ions(i,3)
                     zijsq = zij*zij
                     exp_ij = exp(-zijsq*alphasq)
                     gradmu_efield(jz,iz) = gradmu_efield(jz,iz) + &
                     2.0_wp*volfactor*alpha*sqrpi*exp_ij
                  end do
               end do
            end do
         end do

         ! Blocks on the diagonal compute only half of the pair interactions
         num_block_diag = localwork%pair_ion2ion_num_block_diag_local(itype)
         iblock_offset = localwork%pair_ion2ion_diag_iblock_offset(itype)

         do iblock = 1, num_block_diag
            call update_diag_block_boundaries(iblock_offset+iblock, &
            offset_n, count_n, istart, iend)
            do i = istart, iend
               iz = i + num_ions*2
               do j = istart, iend
                  jz = j + num_ions*2
                  zij = xyz_ions(j,3) - xyz_ions(i,3)
                  zijsq = zij*zij
                  exp_ij = exp(-zijsq*alphasq)
                  gradmu_efield(jz,iz) = gradmu_efield(jz,iz) + &
                  2.0_wp*volfactor*alpha*sqrpi*exp_ij
               end do
            end do
         end do

         ! Number of blocks with full interactions (below the diagonal
         num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,itype)
         iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,itype)
         do iblock = 1, num_block_full
            call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
            istart, iend, jstart, jend)
            do i = istart, iend
               iz = i + num_ions*2
               do j = jstart, jend
                  jz = j + num_ions*2
                  zij = xyz_ions(j,3) - xyz_ions(i,3)
                  zijsq = zij*zij
                  exp_ij = exp(-zijsq*alphasq)
                  gradmu_efield(jz,iz) = gradmu_efield(jz,iz) + &
                  2.0_wp*volfactor*alpha*sqrpi*exp_ij
               end do
            end do
         end do

      end do
   end subroutine gradmumelt_electricfield

   !================================================================================
   ! Compute the Coulomb potential felt by each electrode atoms due to all melt ions
   subroutine mumelt2Qelec_potential(localwork, ewald, box, &
      ions, xyz_ions, electrodes, xyz_atoms, dipolez, V)
      implicit none

      ! Parameters in
      ! -------------
      type(MW_localwork_t), intent(in) :: localwork !< Local work assignment
      type(MW_ewald_t),     intent(in) :: ewald    !< Ewald summation parameters
      type(MW_box_t),       intent(in) :: box      !< Simulation box parameters
      type(MW_ion_t),       intent(in) :: ions(:)        !< ions parameters
      real(wp),             intent(in) :: xyz_ions(:,:)  !< ions xyz positions
      type(MW_electrode_t), intent(in) :: electrodes(:)   !< electrode parameters
      real(wp),             intent(in) :: xyz_atoms(:,:)  !< electrode atoms xyz positions
      real(wp),             intent(in) :: dipolez(:)      !< z-component of the dipole

      ! Parameters out
      ! --------------
      real(wp),                 intent(inout) :: V(:)       !< Coulomb potential

      ! Local
      ! -----
      integer :: num_ion_types, num_elec_types
      integer :: i, j, itype, jtype
      real(wp) :: volfactor
      real(wp) :: zij, zijsq, erf_ij
      real(wp) :: alpha, alphasq, sqrpialpha
      real(wp) :: vi

      integer :: num_block_full
      integer :: iblock, iblock_offset, istart, iend, jstart, jend
      integer :: count_n, offset_n, count_m, offset_m

      num_ion_types = size(ions,1)
      num_elec_types = size(electrodes,1)

      ! Precompute some factors
      ! -----------------------
      alpha = ewald%alpha
      alphasq = alpha * alpha
      sqrpialpha = sqrt(pi) / alpha

      ! Long range contribution from k=0 mode
      ! -------------------------------------
      volfactor = 2.0_wp/box%area(3)

      do itype = 1, num_elec_types
         count_n = electrodes(itype)%count
         offset_n = electrodes(itype)%offset
         do jtype = 1, num_ion_types
            count_m = ions(jtype)%count
            offset_m = ions(jtype)%offset

            ! Number of blocks with full interactions (below the diagonal
            num_block_full = localwork%pair_ion2atom_num_block_full_local(itype,jtype)
            iblock_offset = localwork%pair_ion2atom_full_iblock_offset(itype,jtype)
            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
               offset_n, count_n, offset_m, count_m, &
               istart, iend, jstart, jend)
               do i = istart, iend
                  vi = 0.0_wp
                  do j = jstart, jend
                     zij = xyz_atoms(i,3) - xyz_ions(j,3)
                     zijsq = zij*zij
                     erf_ij = erf(zij*alpha)
                     vi = vi + volfactor*dipolez(j)*pi*erf_ij
                  end do
                  V(i) = V(i) + vi
               end do
            end do
         end do
      end do
   end subroutine mumelt2Qelec_potential

   !================================================================================
   ! Compute the electric fields felt by each melt ions due to other melt ions
   ! charges
   subroutine qmelt2mumelt_electricfield(localwork, ewald, box, ions, xyz_ions, &
      efield)
      implicit none

      ! Parameters in
      ! -------------
      type(MW_localwork_t),  intent(in) :: localwork       !< Local work assignment
      type(MW_ewald_t),      intent(in) :: ewald          !< Ewald summation parameters
      type(MW_box_t),        intent(in) :: box            !< Simulation box parameters
      type(MW_ion_t),        intent(in) :: ions(:)        !< ions parameters
      real(wp),              intent(in) :: xyz_ions(:,:)  !< ions xyz positions

      ! Parameters out
      ! --------------
      real(wp), intent(inout) :: efield(:,:) !< Electric field

      ! Local
      ! -----
      integer :: num_ion_types
      integer :: i, j, itype, jtype
      real(wp) :: qi, qj !< ion charge
      real(wp) :: volfactor
      real(wp) :: zij,  erf_ij
      real(wp) :: alpha

      integer :: num_block_diag, num_block_full
      integer :: iblock, iblock_offset, istart, iend, jstart, jend
      integer :: count_n, offset_n, count_m, offset_m

      num_ion_types = size(ions,1)

      ! Precompute some factors
      ! -----------------------
      alpha = ewald%alpha

      ! Long range contribution from k=0 mode
      ! -------------------------------------
      ! Normalization
      volfactor = 2.0_wp/box%area(3)


      do itype = 1, num_ion_types
         count_n = ions(itype)%count
         offset_n = ions(itype)%offset
         qi = ions(itype)%charge
         do jtype = 1, itype - 1
            count_m = ions(jtype)%count
            offset_m = ions(jtype)%offset
            qj = ions(jtype)%charge
            ! Number of blocks with full ion2ion interactions
            num_block_full = localwork%pair_ion2ion_num_block_full_local(itype, jtype)
            iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype, jtype)

            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
               offset_n, count_n, offset_m, count_m, &
               istart, iend, jstart, jend)
               do i = istart, iend
                  do j = jstart, jend
                     zij = xyz_ions(j,3) - xyz_ions(i,3)
                     erf_ij = erf(alpha*zij)
                     efield(i,3) = efield(i,3) - volfactor*pi*erf_ij*qj
                     efield(j,3) = efield(j,3) + volfactor*pi*erf_ij*qi
                  end do
               end do
            end do
         end do

         ! Blocks on the diagonal compute only half of the pair interactions
         qi = ions(itype)%charge
         qj = ions(itype)%charge
         num_block_diag = localwork%pair_ion2ion_num_block_diag_local(itype)
         iblock_offset = localwork%pair_ion2ion_diag_iblock_offset(itype)

         do iblock = 1, num_block_diag
            call update_diag_block_boundaries(iblock_offset+iblock, &
            offset_n, count_n, istart, iend)
            do i = istart, iend
               do j = istart, iend
                  zij = xyz_ions(j,3) - xyz_ions(i,3)
                  erf_ij = erf(alpha*zij)
                  efield(j,3) = efield(j,3) + volfactor*pi*erf_ij*qi
               end do
            end do
         end do

         ! Number of blocks with full interactions (below the diagonal
         qi = ions(itype)%charge
         qj = ions(itype)%charge
         num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,itype)
         iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,itype)
         do iblock = 1, num_block_full
            call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
            istart, iend, jstart, jend)
            do i = istart, iend
               do j = jstart, jend
                  zij = xyz_ions(j,3) - xyz_ions(i,3)
                  erf_ij = erf(alpha*zij)
                  efield(i,3) = efield(i,3) - volfactor*pi*erf_ij*qj
                  efield(j,3) = efield(j,3) + volfactor*pi*erf_ij*qi
               end do
            end do
         end do
      end do
   end subroutine qmelt2mumelt_electricfield

   !================================================================================
   ! Compute the electric fields felt by each melt ions due to other melt ions
   ! dipoles
   subroutine mumelt2mumelt_electricfield(localwork, ewald, box, ions, xyz_ions, &
      efield, dipolez)
      implicit none

      ! Parameters in
      ! -------------
      type(MW_localwork_t),  intent(in) :: localwork       !< Local work assignment
      type(MW_ewald_t),     intent(in) :: ewald          !< Ewald summation parameters
      type(MW_box_t),       intent(in) :: box            !< Simulation box parameters
      type(MW_ion_t),       intent(in) :: ions(:)        !< ions parameters
      real(wp),             intent(in) :: xyz_ions(:,:)  !< ions xyz positions
      real(wp),             intent(in) :: dipolez(:)      !< z-component of the dipole

      ! Parameters out
      ! --------------
      real(wp), intent(inout) :: efield(:,:) !< Electric field on ions

      ! Local
      ! -----
      integer :: num_ion_types
      integer :: i, j, itype, jtype
      real(wp) :: muiz,mujz
      real(wp) :: volfactor
      real(wp) :: zij, zijsq, exp_ij
      real(wp) :: alpha, alphasq, alphapi, sqrpialpha,sqrpi
      integer :: num_block_diag, num_block_full
      integer :: iblock, iblock_offset, istart, iend, jstart, jend
      integer :: count_n, offset_n, count_m, offset_m

      num_ion_types = size(ions,1)

      ! Precompute some factors
      ! -----------------------
      alpha = ewald%alpha
      alphasq = alpha*alpha
      sqrpi=sqrt(pi)
      alphapi = 2.0_wp * alpha / sqrpi
      sqrpialpha = sqrpi / alpha

      ! Long range contribution from k=0 mode
      ! -------------------------------------
      ! Normalization
      volfactor = 2.0_wp/box%area(3)


      do itype = 1, num_ion_types
         count_n = ions(itype)%count
         offset_n = ions(itype)%offset
         do jtype = 1, itype - 1
            count_m = ions(jtype)%count
            offset_m = ions(jtype)%offset
            ! Number of blocks with full ion2ion interactions
            num_block_full = localwork%pair_ion2ion_num_block_full_local(itype, jtype)
            iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype, jtype)

            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
               offset_n, count_n, offset_m, count_m, &
               istart, iend, jstart, jend)
               do i = istart, iend
                  muiz = dipolez(i)
                  do j = jstart, jend
                     mujz = dipolez(j)
                     zij = xyz_ions(j,3) - xyz_ions(i,3)
                     zijsq = zij*zij
                     exp_ij = exp(-zijsq*alphasq)
                     efield(i,3) = efield(i,3) - 2.0_wp*volfactor*alpha*sqrpi*exp_ij*mujz
                     efield(j,3) = efield(j,3) - 2.0_wp*volfactor*alpha*sqrpi*exp_ij*muiz
                  end do
               end do
            end do
         end do

         ! Blocks on the diagonal compute only half of the pair interactions
         num_block_diag = localwork%pair_ion2ion_num_block_diag_local(itype)
         iblock_offset = localwork%pair_ion2ion_diag_iblock_offset(itype)

         do iblock = 1, num_block_diag
            call update_diag_block_boundaries(iblock_offset+iblock, &
            offset_n, count_n, istart, iend)
            do i = istart, iend
               muiz = dipolez(i)
               do j = istart, iend
                  mujz = dipolez(j)
                  zij = xyz_ions(j,3) - xyz_ions(i,3)
                  zijsq = zij*zij
                  exp_ij = exp(-zijsq*alphasq)
                  efield(j,3) = efield(j,3) - 2.0_wp*volfactor*alpha*sqrpi*exp_ij*muiz
               end do
            end do
         end do

         ! Number of blocks with full interactions (below the diagonal
         num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,itype)
         iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,itype)
         do iblock = 1, num_block_full
            call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
            istart, iend, jstart, jend)
            do i = istart, iend
               muiz = dipolez(i)
               do j = jstart, jend
                  mujz = dipolez(j)
                  zij = xyz_ions(j,3) - xyz_ions(i,3)
                  zijsq = zij*zij
                  exp_ij = exp(-zijsq*alphasq)
                  efield(i,3) = efield(i,3) - 2.0_wp*volfactor*alpha*sqrpi*exp_ij*mujz
                  efield(j,3) = efield(j,3) - 2.0_wp*volfactor*alpha*sqrpi*exp_ij*muiz
               end do
            end do
         end do
      end do
   end subroutine mumelt2mumelt_electricfield

   !================================================================================
   ! Compute the electric fields felt by each melt ions due to other melt ions
   ! charges
   subroutine Qelec2mumelt_electricfield(localwork, ewald, box, ions, xyz_ions, &
      electrodes, xyz_atoms, q_elec, efield)
      implicit none

      ! Parameters in
      ! -------------
      type(MW_localwork_t),  intent(in) :: localwork       !< Local work assignment
      type(MW_ewald_t),     intent(in) :: ewald          !< Ewald summation parameters
      type(MW_box_t),       intent(in) :: box            !< Simulation box parameters
      type(MW_ion_t),       intent(in) :: ions(:)        !< ions parameters
      real(wp),             intent(in) :: xyz_ions(:,:)  !< ions xyz positions
      type(MW_electrode_t), intent(in) :: electrodes(:)   !< electrode parameters
      real(wp),             intent(in) :: xyz_atoms(:,:)  !< electrode atoms xyz positions
      real(wp),             intent(in) :: q_elec(:)      !< Electrode atoms charge

      ! Parameters out
      ! --------------
      real(wp), intent(inout) :: efield(:,:) !< Electric field

      ! Local
      ! -----
      integer :: num_ion_types, num_elec_types
      integer :: i, j, itype, jtype
      real(wp) :: qi !< ion charge
      real(wp) :: volfactor
      real(wp) :: zij,  erf_ij
      real(wp) :: alpha
      integer :: num_block_full
      integer :: iblock, iblock_offset, istart, iend, jstart, jend
      integer :: count_n, offset_n, count_m, offset_m

      num_ion_types = size(ions,1)
      num_elec_types = size(electrodes,1)

      ! Precompute some factors
      ! -----------------------
      alpha = ewald%alpha

      ! Long range contribution from k=0 mode
      ! -------------------------------------
      ! Normalization
      volfactor = 2.0_wp/box%area(3)


      do itype = 1, num_ion_types
         count_n = ions(itype)%count
         offset_n = ions(itype)%offset
         qi = ions(itype)%charge
         do jtype = 1, num_elec_types
            count_m = electrodes(jtype)%count
            offset_m = electrodes(jtype)%offset
            ! Number of blocks with full atom2ion interactions
            num_block_full = localwork%pair_atom2ion_num_block_full_local(itype, jtype)
            iblock_offset = localwork%pair_atom2ion_full_iblock_offset(itype, jtype)
            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
               offset_n, count_n, offset_m, count_m, &
               istart, iend, jstart, jend)
               do i = istart, iend
                  do j = jstart, jend
                     zij = xyz_atoms(j,3) - xyz_ions(i,3)
                     erf_ij = erf(alpha*zij)
                     efield(i,3) = efield(i,3) - volfactor*pi*erf_ij*q_elec(j)
                  end do
               end do
            end do
         end do
      end do
   end subroutine Qelec2mumelt_electricfield

   !================================================================================
   ! Compute the Coulomb forces felt by each melt ions due to other melt ions
   subroutine melt_forces(localwork, ewald, box, ions, xyz_ions, &
      electrodes, xyz_atoms, q_elec, force, dipolez)
      implicit none

      ! Parameters in
      ! -------------
      type(MW_localwork_t),  intent(in) :: localwork       !< Local work assignment
      type(MW_ewald_t),     intent(in) :: ewald          !< Ewald summation parameters
      type(MW_box_t),       intent(in) :: box            !< Simulation box parameters
      type(MW_ion_t),       intent(in) :: ions(:)        !< ions parameters
      real(wp),             intent(in) :: xyz_ions(:,:)  !< ions xyz positions
      type(MW_electrode_t), intent(in) :: electrodes(:)   !< electrode parameters
      real(wp),             intent(in) :: xyz_atoms(:,:)  !< electrode atoms xyz positions
      real(wp),             intent(in) :: q_elec(:)      !< Electrode atoms charge
      real(wp),             intent(in) :: dipolez(:)      !< z-component of the dipole

      ! Parameters out
      ! --------------
      real(wp), intent(inout) :: force(:,:) !< Coulomb force on ions

      ! Local
      ! -----
      integer :: num_ion_types, num_elec_types
      integer :: i, j, itype, jtype
      real(wp) :: qi, qj !< ion charge
      real(wp) :: Q_j    !< electrode charge
      real(wp) :: muiz,mujz
      real(wp) :: volfactor
      real(wp) :: zij, zijsq, erf_ij,exp_ij
      real(wp) :: fij
      real(wp) :: alpha, alphasq, alphapi, sqrpialpha,sqrpi
      integer :: num_block_diag, num_block_full
      integer :: iblock, iblock_offset, istart, iend, jstart, jend
      integer :: count_n, offset_n, count_m, offset_m

      num_ion_types = size(ions,1)
      num_elec_types = size(electrodes,1)

      ! Precompute some factors
      ! -----------------------
      alpha = ewald%alpha
      alphasq = alpha*alpha
      sqrpi = sqrt(pi)
      alphapi = 2.0_wp * alpha / sqrpi
      sqrpialpha = sqrpi / alpha

      ! Long range contribution from k=0 mode
      ! -------------------------------------
      ! Normalization
      volfactor = 2.0_wp/box%area(3)


      do itype = 1, num_ion_types
         count_n = ions(itype)%count
         offset_n = ions(itype)%offset
         do jtype = 1, itype - 1
            count_m = ions(jtype)%count
            offset_m = ions(jtype)%offset
            qi = ions(itype)%charge
            qj = ions(jtype)%charge
            ! Number of blocks with full ion2ion interactions
            num_block_full = localwork%pair_ion2ion_num_block_full_local(itype, jtype)
            iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype, jtype)

            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
               offset_n, count_n, offset_m, count_m, &
               istart, iend, jstart, jend)
               do i = istart, iend
                  muiz = dipolez(i)
                  do j = jstart, jend
                     mujz = dipolez(j)
                     zij = xyz_ions(j,3) - xyz_ions(i,3)
                     zijsq = zij*zij
                     erf_ij = erf(alpha*zij)
                     exp_ij = exp(-zijsq*alphasq)
                     fij = qi * qj * pi * volfactor * erf_ij   &
                     +2.0_wp*volfactor*exp_ij*sqrpi*alpha*(qj*muiz - qi*mujz)  &
                     -4.0_wp*volfactor*zij*alphasq*alpha*sqrpi*muiz*mujz*exp_ij
                     force(i,3) = force(i,3) - fij
                     force(j,3) = force(j,3) + fij
                  end do
               end do
            end do
         end do

         ! Blocks on the diagonal compute only half of the pair interactions
         qi = ions(itype)%charge
         qj = ions(itype)%charge
         num_block_diag = localwork%pair_ion2ion_num_block_diag_local(itype)
         iblock_offset = localwork%pair_ion2ion_diag_iblock_offset(itype)

         do iblock = 1, num_block_diag
            call update_diag_block_boundaries(iblock_offset+iblock, &
            offset_n, count_n, istart, iend)
            do i = istart, iend
               muiz = dipolez(i)
               do j = istart, iend
                  mujz = dipolez(j)
                  zij = xyz_ions(j,3) - xyz_ions(i,3)
                  zijsq = zij*zij
                  erf_ij = erf(alpha*zij)
                  exp_ij = exp(-zijsq*alphasq)
                  fij = qi * qj * pi * volfactor * erf_ij  &
                  +2.0_wp*volfactor*exp_ij*sqrpi*alpha*(qj*muiz - qi*mujz) &
                  -4.0_wp*volfactor*zij*alphasq*alpha*sqrpi*muiz*mujz*exp_ij
                  force(j,3) = force(j,3) + fij
               end do
            end do
         end do

         ! Number of blocks with full interactions (below the diagonal
         qi = ions(itype)%charge
         qj = ions(itype)%charge
         num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,itype)
         iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,itype)
         do iblock = 1, num_block_full
            call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
            istart, iend, jstart, jend)
            do i = istart, iend
               muiz = dipolez(i)
               do j = jstart, jend
                  mujz = dipolez(j)
                  zij = xyz_ions(j,3) - xyz_ions(i,3)
                  zijsq = zij*zij
                  erf_ij = erf(alpha*zij)
                  exp_ij = exp(-zijsq*alphasq)
                  fij = qi * qj * pi * volfactor * erf_ij &
                  +2.0_wp*volfactor*exp_ij*sqrpi*alpha*(qj*muiz - qi*mujz) &
                  -4.0_wp*volfactor*zij*alphasq*alpha*sqrpi*muiz*mujz*exp_ij
                  force(j,3) = force(j,3) + fij
                  force(i,3) = force(i,3) - fij
               end do
            end do
         end do

         qi = ions(itype)%charge
         do jtype = 1, num_elec_types
            count_m = electrodes(jtype)%count
            offset_m = electrodes(jtype)%offset
            ! Number of blocks with full atom2ion interactions
            num_block_full = localwork%pair_atom2ion_num_block_full_local(itype, jtype)
            iblock_offset = localwork%pair_atom2ion_full_iblock_offset(itype, jtype)
            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
               offset_n, count_n, offset_m, count_m, &
               istart, iend, jstart, jend)
               do i = istart, iend
                  muiz = dipolez(i)
                  do j = jstart, jend
                     Q_j = q_elec(j)
                     zij = xyz_atoms(j,3) - xyz_ions(i,3)
                     zijsq = zij*zij
                     erf_ij = erf(alpha*zij)
                     exp_ij = exp(-zijsq*alphasq)
                     fij = qi*Q_j*pi*volfactor*erf_ij &
                     + 2.0_wp*volfactor*exp_ij*sqrpi*alpha*(Q_j*muiz)
                     force(i,3) = force(i,3) - fij
                  end do
               end do
            end do
         end do
      end do
   end subroutine melt_forces

   !================================================================================
   ! Compute the contibution to the energy from long-range Coulomb interaction
   subroutine energy(localwork, ewald, box, ions, xyz_ions, &
      electrodes, xyz_atoms, q_elec, h, dipolez)
      implicit none

      ! Parameters in
      ! -------------
      type(MW_localwork_t), intent(in) :: localwork       !< Local work assignment
      type(MW_ewald_t),     intent(in) :: ewald          !< Ewald summation parameters
      type(MW_box_t),       intent(in) :: box            !< Simulation box parameters
      type(MW_ion_t),       intent(in) :: ions(:)        !< ions parameters
      real(wp),             intent(in) :: xyz_ions(:,:)  !< ions xyz positions
      type(MW_electrode_t), intent(in) :: electrodes(:)   !< electrode parameters
      real(wp),             intent(in) :: xyz_atoms(:,:)  !< electrode atoms xyz positions
      real(wp),             intent(in) :: q_elec(:)      !< Electrode atoms charge
      real(wp),             intent(in) :: dipolez(:)      !< z-component of the dipole

      ! Parameters out
      ! --------------
      real(wp), intent(inout) :: h !< Energy

      ! Local
      ! -----
      integer :: num_ion_types, num_elec_types
      integer :: i, j, itype, jtype
      real(wp) :: qi, qj !< ion charge
      real(wp) :: Q_j !< electrode charge
      real(wp) :: muiz, mujz
      real(wp) :: volfactor
      real(wp) :: zij, zijsq
      real(wp) :: erf_ij,exp_ij
      real(wp) :: pot_ij, vsumzk0
      real(wp) :: alpha, alphasq, alphapi, sqrpialpha,sqrpi

      integer :: num_block_diag, num_block_full
      integer :: iblock, iblock_offset, istart, iend, jstart, jend
      integer :: count_n, offset_n, count_m, offset_m

      num_ion_types = size(ions,1)
      num_elec_types = size(electrodes,1)

      ! Precompute some factors
      ! -----------------------
      alpha = ewald%alpha
      alphasq = alpha*alpha
      sqrpi = sqrt(pi)
      alphapi = 2.0_wp*alpha/sqrpi
      sqrpialpha = sqrpi/alpha

      ! Zero energy
      ! ----------------
      h = 0.0_wp
      vsumzk0 = 0.0_wp
      ! Long range contribution from k=0 mode
      ! -------------------------------------
      ! Normalization
      volfactor = 2.0_wp/box%area(3)


      ! Melt -> Melt and Elec -> Melt interactions
      ! ------------------------------------------
      do itype = 1, num_ion_types
         count_n = ions(itype)%count
         offset_n = ions(itype)%offset
         qi = ions(itype)%charge
         do jtype = 1, itype - 1
            count_m = ions(jtype)%count
            offset_m = ions(jtype)%offset
            qj = ions(jtype)%charge
            ! Number of blocks with full ion2ion interactions
            num_block_full = localwork%pair_ion2ion_num_block_full_local(itype, jtype)
            iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype, jtype)

            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock+iblock_offset, &
               offset_n, count_n, offset_m, count_m, &
               istart, iend, jstart, jend)

               do i = istart, iend
                  muiz = dipolez(i)
                  do j = jstart, jend
                     mujz = dipolez(j)
                     zij = xyz_ions(j,3) - xyz_ions(i,3)
                     zijsq = zij*zij
                     erf_ij = erf(zij*alpha)
                     exp_ij = exp(-zijsq*alphasq)
                     pot_ij = (sqrpialpha*exp_ij + pi*zij*erf_ij)
                     ! The minus sign for qmu term arises from the different definition of zij
                     ! w.r.t. Pounds PhD (erf is an impair function)
                     vsumzk0 = vsumzk0 + qi * qj * pot_ij &
                     -pi*(qj*muiz - qi*mujz)*erf_ij   &
                     -2.0_wp*muiz*mujz*alpha*sqrpi*exp_ij
                  end do
               end do
            end do
         end do

         ! Blocks on the diagonal compute only half of the pair interactions
         qi = ions(itype)%charge
         qj = ions(itype)%charge
         num_block_diag = localwork%pair_ion2ion_num_block_diag_local(itype)
         iblock_offset = localwork%pair_ion2ion_diag_iblock_offset(itype)

         do iblock = 1, num_block_diag
            call update_diag_block_boundaries(iblock_offset+iblock, offset_n, count_n, istart, iend)
            do i = istart, iend
               muiz = dipolez(i)
               do j = istart, iend
                  mujz = dipolez(j)
                  zij = xyz_ions(j,3) - xyz_ions(i,3)
                  zijsq = zij*zij
                  erf_ij = erf(zij*alpha)
                  exp_ij = exp(-zijsq*alphasq)
                  pot_ij = (sqrpialpha*exp_ij + pi*zij*erf_ij)
                  ! The minus sign for qmu term arises from the different definition of zij
                  ! w.r.t. Pounds PhD (erf is an impair function)
                  vsumzk0 = vsumzk0 + 0.5_wp*(qi * qj * pot_ij &
                  -pi*(qj*muiz - qi*mujz)*erf_ij   &
                  -2.0_wp*muiz*mujz*alpha*sqrpi*exp_ij)
               end do
            end do
         end do

         ! Number of blocks with full interactions (below the diagonal
         qi = ions(itype)%charge
         qj = ions(itype)%charge
         num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,itype)
         iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,itype)

         do iblock = 1, num_block_full
            call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
            istart, iend, jstart, jend)
            do i = istart, iend
               muiz = dipolez(i)
               do j = jstart, jend
                  mujz = dipolez(j)
                  zij = xyz_ions(j,3) - xyz_ions(i,3)
                  zijsq = zij*zij
                  erf_ij = erf(zij*alpha)
                  exp_ij = exp(-zijsq*alphasq)
                  pot_ij = (sqrpialpha*exp_ij + pi*zij*erf_ij)
                  ! The minus sign for qmu term arises from the different definition of zij
                  ! w.r.t. Pounds PhD (erf is an impair function)
                  vsumzk0 = vsumzk0 + qi * qj * pot_ij &
                        -pi*(qj*muiz - qi*mujz)*erf_ij   &
                        -2.0_wp*muiz*mujz*alpha*sqrpi*exp_ij
               end do
            end do
         end do

         !qi -> point charge (melt)
         !muiz -> z-component of dipoles (melt)
         !Q_j -> integrated gaussian charge (walls)
         qi = ions(itype)%charge
         do jtype = 1, num_elec_types
            count_m = electrodes(jtype)%count
            offset_m = electrodes(jtype)%offset
            ! Number of blocks with full ion2ion interactions
            num_block_full = localwork%pair_atom2ion_num_block_full_local(itype, jtype)
            iblock_offset = localwork%pair_atom2ion_full_iblock_offset(itype, jtype)

            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock+iblock_offset, &
               offset_n, count_n, offset_m, count_m, &
               istart, iend, jstart, jend)
               do i = istart, iend
                  muiz = dipolez(i)
                  do j = jstart, jend
                     Q_j = q_elec(j)
                     zij = xyz_atoms(j,3) - xyz_ions(i,3)
                     zijsq = zij*zij
                     erf_ij = erf(zij*alpha)
                     exp_ij = exp(-zijsq*alphasq)
                     pot_ij = (sqrpialpha*exp_ij + pi*zij*erf_ij)
                     vsumzk0 = vsumzk0 + qi*Q_j*pot_ij &
                           -pi*Q_j*muiz*erf_ij
                  end do
               end do
            end do
         end do
      end do

      ! Elec 2 Elec interactions
      ! ------------------------
      ! Blocks on the diagonal compute only half of the pair interactions
      do itype = 1, num_elec_types
         count_n = electrodes(itype)%count
         offset_n = electrodes(itype)%offset
         do jtype = 1, itype-1
            count_m =  electrodes(jtype)%count
            offset_m = electrodes(jtype)%offset

            ! Number of blocks with full interactions (below the diagonal
            num_block_full = localwork%pair_atom2atom_num_block_full_local(itype,jtype)
            iblock_offset = localwork%pair_atom2atom_full_iblock_offset(itype,jtype)

            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
               offset_n, count_n, offset_m, count_m, istart, iend, jstart, jend)
               do i = istart, iend
                  do j = jstart, jend
                     zij = xyz_atoms(j,3) - xyz_atoms(i,3)
                     zijsq = zij*zij
                     pot_ij = (sqrpialpha*exp(-zijsq*alphasq) + pi*zij*erf(zij*alpha))
                     vsumzk0 = vsumzk0 + q_elec(i) * q_elec(j) * pot_ij
                  end do
               end do
            end do
         end do

         num_block_diag = localwork%pair_atom2atom_num_block_diag_local(itype)
         iblock_offset = localwork%pair_atom2atom_diag_iblock_offset(itype)

         do iblock = 1, num_block_diag
            call update_diag_block_boundaries(iblock+iblock_offset, &
            offset_n, count_n, istart, iend)
            do i = istart, iend
               do j = istart, iend
                  zij = xyz_atoms(j,3) - xyz_atoms(i,3)
                  zijsq = zij*zij
                  pot_ij = (sqrpialpha*exp(-zijsq*alphasq) + pi*zij*erf(zij*alpha))
                  vsumzk0 = vsumzk0 + 0.5_wp * q_elec(i) * q_elec(j) * pot_ij
               end do
            end do
         end do

         ! Number of blocks with full interactions (below the diagonal
         num_block_full = localwork%pair_atom2atom_num_block_full_local(itype,itype)
         iblock_offset = localwork%pair_atom2atom_full_iblock_offset(itype,itype)

         do iblock = 1, num_block_full

            call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
            istart, iend, jstart, jend)
            do i = istart, iend
               do j = jstart, jend
                  zij = xyz_atoms(j,3) - xyz_atoms(i,3)
                  zijsq = zij*zij
                  pot_ij = (sqrpialpha*exp(-zijsq*alphasq) + pi*zij*erf(zij*alpha))
                  vsumzk0 = vsumzk0 + q_elec(i) * q_elec(j) * pot_ij
               end do
            end do
         end do
      end do

      ! scale by volume factor
      h = -vsumzk0 * volfactor

   end subroutine energy

   include 'update_diag_block_boundaries.inc'
   include 'update_tri_block_boundaries.inc'
   include 'update_other_block_boundaries.inc'
end module MW_coulomb_keq0_pim
