!================================================================================
! Compute the Coulomb potential felt by each electrode atoms due to all electrode atoms
!DEC$ ATTRIBUTES NOINLINE :: elec2elec_potential_2DPBC
subroutine gradQelec_potential_@PBC@(localwork, ewald, box, &
      electrodes, xyz_atoms, gradQ_V)
   implicit none

   ! Parameters in
   ! -------------
   type(MW_localwork_t), intent(in)    :: localwork  !< Localwork work distribution
   type(MW_ewald_t),     intent(in)    :: ewald     !< Ewald summation parameters
   type(MW_box_t),       intent(in)    :: box       !< Simulation box parameters
   type(MW_electrode_t), intent(in)    :: electrodes(:)  !< electrode parameters
   real(wp),             intent(in)    :: xyz_atoms(:,:)  !< Electrode atoms positions

   ! Parameters out
   ! --------------
   real(wp),             intent(inout) :: gradQ_V(:,:)       !< Coulomb potential

   ! Local
   ! -----
   integer :: num_elec_types
   integer :: i, j
   real(wp) :: a, b, c
   real(wp) :: pot_ij
   real(wp) :: drnorm, drnorm2
   real(wp) :: etai, etaj, etaij, etaijsq, erfdiff
   real(wp) :: alpha, alphasq

   integer :: iblock, iblock_offset, istart, iend, jstart, jend, itype, jtype
   integer :: num_block_diag, num_block_full
   integer :: count_n, offset_n, count_m, offset_m

   num_elec_types = size(electrodes,1)

   ! Precompute some factors
   ! -----------------------
   a = box%length(1)
   b = box%length(2)
   c = box%length(3)

   alpha = ewald%alpha
   alphasq = alpha*alpha


   ! Short range contribution
   ! ------------------------
   do itype = 1, num_elec_types
      count_n = electrodes(itype)%count
      offset_n = electrodes(itype)%offset
      do jtype = 1, itype-1
         count_m = electrodes(jtype)%count
         offset_m = electrodes(jtype)%offset
         if (ewald%sr_interaction_atom2atom(itype,jtype)) then
            etai = electrodes(itype)%eta
            etaj = electrodes(jtype)%eta
            etaij = (etai*etaj)/sqrt(etai*etai + etaj*etaj)
            etaijsq = etaij*etaij

            ! Number of blocks with full interactions (below the diagonal
            num_block_full = localwork%pair_atom2atom_num_block_full_local(itype,jtype)
            iblock_offset = localwork%pair_atom2atom_full_iblock_offset(itype,jtype)
            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
                     offset_n, count_n, offset_m, count_m, &
                     istart, iend, jstart, jend)
               do i = istart, iend
                  do j = jstart, jend
                     call minimum_image_distance_2DPBC(a, b, c, &
                           xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                           xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                           drnorm2)
                     if (drnorm2 < ewald%rcutsq)  then
                        drnorm = sqrt(drnorm2)
                        ! Change erfc to 1-erf because of underflows
                        ! erfdiff = exp(-alphasq*drnorm2) * &
                        !       (erfc_scaled(alpha*drnorm) - exp((alphasq - etaijsq)*drnorm2) * erfc_scaled(etaij*drnorm))
                        ! erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                        erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                        pot_ij =  erfdiff/drnorm
                        gradQ_V(j,i) = gradQ_V(j,i) + pot_ij
                     end if
                  end do
               end do
            end do
         end if
      end do

      if (ewald%sr_interaction_atom2atom(itype,itype)) then

         etai = electrodes(itype)%eta
         etaij = etai/sqrt(2.0_wp)
         etaijsq = etaij*etaij

         ! Blocks on the diagonal compute only half of the pair interactions
         num_block_diag = localwork%pair_atom2atom_num_block_diag_local(itype)
         iblock_offset = localwork%pair_atom2atom_diag_iblock_offset(itype)
         do iblock = 1, num_block_diag
            call update_diag_block_boundaries(iblock_offset+iblock, offset_n, count_n, istart, iend)
            do i = istart, iend
               do j = istart, iend
                  call minimum_image_distance_2DPBC(a, b, c, &
                        xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                        xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                        drnorm2)
                  if ((drnorm2 > 0.0_wp) .and. (drnorm2 < ewald%rcutsq)) then
                     drnorm = sqrt(drnorm2)
                     erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                     pot_ij =  erfdiff/drnorm
                     gradQ_V(j,i) = gradQ_V(j,i) + pot_ij
                  end if
               end do
            end do
         end do

         ! Number of blocks with full interactions (below the diagonal
         num_block_full = localwork%pair_atom2atom_num_block_full_local(itype,itype)
         iblock_offset = localwork%pair_atom2atom_full_iblock_offset(itype,itype)
         do iblock = 1, num_block_full
            call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
                  istart, iend, jstart, jend)
            do i = istart, iend
               do j = jstart, jend
                  call minimum_image_distance_2DPBC(a, b, c, &
                        xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                        xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                        drnorm2)
                  if (drnorm2 < ewald%rcutsq)  then
                     drnorm = sqrt(drnorm2)
                     erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                     pot_ij =  erfdiff/drnorm
                     gradQ_V(j,i) = gradQ_V(j,i) + pot_ij
                  end if
               end do
            end do
         end do
      end if
   end do
end subroutine gradQelec_potential_@PBC@

!================================================================================
! Compute the Coulomb potential felt by each electrode atoms due to all melt ions
!DEC$ ATTRIBUTES NOINLINE :: melt2elec_potential_@PBC@
subroutine qmelt2Qelec_potential_@PBC@(localwork, ewald, box, &
      ions, xyz_ions, electrodes, xyz_atoms, V)
   implicit none

   ! Parameters in
   ! -------------
   type(MW_localwork_t), intent(in)    :: localwork !< Local work assignment
   type(MW_ewald_t),     intent(in)    :: ewald     !< Ewald summation parameters
   type(MW_box_t),       intent(in)    :: box       !< Simulation box parameters
   type(MW_ion_t),       intent(in)    :: ions(:)        !< ions parameters
   real(wp),             intent(in)    :: xyz_ions(:,:)  !< ions xyz positions
   type(MW_electrode_t), intent(in)    :: electrodes(:)   !< electrode parameters
   real(wp),             intent(in)    :: xyz_atoms(:,:)  !< electrode atoms xyz positions

   ! Parameters out
   ! --------------
   real(wp),             intent(inout) :: V(:)       !< Coulomb potential

   ! Local
   ! -----
   integer :: num_ion_types, num_elec_types
   integer :: i, j, itype, jtype
   real(wp) :: a, b, c
   real(wp) :: qj !< ion charge
   real(wp) :: pot_ij, vi
   real(wp) :: drnorm, drnorm2, drnormrec
   real(wp) :: etai, etaisq, erfdiff
   real(wp) :: alpha, alphasq
   integer :: num_block_full
   integer :: iblock, iblock_offset, istart, iend, jstart, jend
   integer :: count_n, offset_n, count_m, offset_m

   num_ion_types = size(ions,1)
   num_elec_types = size(electrodes,1)

   ! Precompute some factors
   ! -----------------------
   a = box%length(1)
   b = box%length(2)
   c = box%length(3)

   alpha = ewald%alpha
   alphasq = alpha*alpha


   ! Short range contribution
   ! ------------------------
   do itype = 1, num_elec_types
      count_n = electrodes(itype)%count
      offset_n = electrodes(itype)%offset

      etai = electrodes(itype)%eta
      etaisq = etai*etai
      do jtype = 1, num_ion_types
         count_m = ions(jtype)%count
         offset_m = ions(jtype)%offset

         qj = ions(jtype)%charge
         if ((qj /= 0.0_wp) .and. ewald%sr_interaction_ion2atom(itype,jtype)) then
            ! Number of blocks with full interactions (below the diagonal
            num_block_full = localwork%pair_ion2atom_num_block_full_local(itype,jtype)
            iblock_offset = localwork%pair_ion2atom_full_iblock_offset(itype,jtype)
            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
                     offset_n, count_n, offset_m, count_m, &
                     istart, iend, jstart, jend)
               do i = istart, iend
                  vi = 0.0_wp
                  do j = jstart, jend
                     call minimum_image_distance_@PBC@(a, b, c, &
                           xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                           xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                           drnorm2)
                     if (drnorm2 < ewald%rcutsq) then
                        drnorm = sqrt(drnorm2)
                        drnormrec = 1.0_wp/drnorm
                        erfdiff = (erfc(alpha*drnorm) - erfc(etai*drnorm))
                        pot_ij =  erfdiff/drnorm
                        Vi = Vi + qj*pot_ij
                     end if
                  end do
                  V(i) = V(i) + vi
               end do
            end do
         end if
      end do
   end do
end subroutine qmelt2Qelec_potential_@PBC@

!================================================================================
! Compute the Coulomb potential felt by each electrode atoms due to all electrode atoms
!DEC$ ATTRIBUTES NOINLINE :: elec2elec_potential_@PBC@
subroutine Qelec2Qelec_potential_@PBC@(localwork, ewald, box, &
      electrodes, xyz_atoms, q_elec, V)
   implicit none

   ! Parameters in
   ! -------------
   type(MW_localwork_t), intent(in)    :: localwork  !< Localwork work distribution
   type(MW_ewald_t),     intent(in)    :: ewald     !< Ewald summation parameters
   type(MW_box_t),       intent(in)    :: box       !< Simulation box parameters
   type(MW_electrode_t), intent(in)    :: electrodes(:)  !< electrode parameters
   real(wp),             intent(in)    :: q_elec(:)      !< Electrode atoms charge
   real(wp),             intent(in)    :: xyz_atoms(:,:)  !< Electrode atoms positions

   ! Parameters out
   ! --------------
   real(wp),             intent(inout) :: V(:)       !< Coulomb potential

   ! Local
   ! -----
   integer :: num_elec_types
   integer :: i, j
   real(wp) :: a, b, c
   real(wp) :: pot_ij, vi, qi
   real(wp) :: drnorm, drnorm2
   real(wp) :: etai, etaj, etaij, etaijsq, erfdiff
   real(wp) :: alpha, alphasq
   integer :: iblock, iblock_offset, istart, iend, jstart, jend, itype, jtype
   integer :: num_block_diag, num_block_full
   integer :: count_n, offset_n, count_m, offset_m

   num_elec_types = size(electrodes,1)

   ! Precompute some factors
   ! -----------------------
   a = box%length(1)
   b = box%length(2)
   c = box%length(3)

   alpha = ewald%alpha
   alphasq = alpha*alpha


   ! Short range contribution
   ! ------------------------
   do itype = 1, num_elec_types
      count_n = electrodes(itype)%count
      offset_n = electrodes(itype)%offset
      do jtype = 1, itype-1
         count_m = electrodes(jtype)%count
         offset_m = electrodes(jtype)%offset
         if (ewald%sr_interaction_atom2atom(itype,jtype)) then
            etai = electrodes(itype)%eta
            etaj = electrodes(jtype)%eta
            etaij = (etai*etaj)/sqrt(etai*etai + etaj*etaj)
            etaijsq = etaij*etaij

            ! Number of blocks with full interactions (below the diagonal
            num_block_full = localwork%pair_atom2atom_num_block_full_local(itype,jtype)
            iblock_offset = localwork%pair_atom2atom_full_iblock_offset(itype,jtype)
            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
                     offset_n, count_n, offset_m, count_m, &
                     istart, iend, jstart, jend)
               do i = istart, iend
                  vi = 0.0_wp
                  qi = q_elec(i)
                  do j = jstart, jend
                     call minimum_image_distance_@PBC@(a, b, c, &
                           xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                           xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                           drnorm2)
                     if (drnorm2 < ewald%rcutsq)  then
                        drnorm = sqrt(drnorm2)
                        ! Change erfc to 1-erf because of underflows
                        ! erfdiff = exp(-alphasq*drnorm2) * &
                        !       (erfc_scaled(alpha*drnorm) - exp((alphasq - etaijsq)*drnorm2) * erfc_scaled(etaij*drnorm))
                        ! erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                        erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                        pot_ij =  erfdiff/drnorm
                        V(j) = V(j) + qi*pot_ij
                        vi = vi + q_elec(j)*pot_ij
                     end if
                  end do
                  V(i) = V(i) + vi
               end do
            end do
         end if
      end do

      if (ewald%sr_interaction_atom2atom(itype,itype)) then

         etai = electrodes(itype)%eta
         etaij = etai/sqrt(2.0_wp)
         etaijsq = etaij*etaij

         ! Blocks on the diagonal compute only half of the pair interactions
         num_block_diag = localwork%pair_atom2atom_num_block_diag_local(itype)
         iblock_offset = localwork%pair_atom2atom_diag_iblock_offset(itype)
         do iblock = 1, num_block_diag
            call update_diag_block_boundaries(iblock_offset+iblock, offset_n, count_n, istart, iend)
            do i = istart, iend
               vi = 0.0_wp
               do j = istart, iend
                  call minimum_image_distance_@PBC@(a, b, c, &
                        xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                        xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                        drnorm2)
                  if ((drnorm2 > 0.0_wp) .and. (drnorm2 < ewald%rcutsq)) then
                     drnorm = sqrt(drnorm2)
                     erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                     pot_ij =  erfdiff/drnorm
                     vi = vi + q_elec(j)*pot_ij
                  end if
               end do
               V(i) = V(i) + vi
            end do
         end do

         ! Number of blocks with full interactions (below the diagonal
         num_block_full = localwork%pair_atom2atom_num_block_full_local(itype,itype)
         iblock_offset = localwork%pair_atom2atom_full_iblock_offset(itype,itype)
         do iblock = 1, num_block_full
            call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
                  istart, iend, jstart, jend)
            do i = istart, iend
               vi = 0.0_wp
               qi = q_elec(i)
               do j = jstart, jend
                  call minimum_image_distance_@PBC@(a, b, c, &
                        xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                        xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                        drnorm2)
                  if (drnorm2 < ewald%rcutsq)  then
                     drnorm = sqrt(drnorm2)
                     erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                     pot_ij =  erfdiff/drnorm
                     V(j) = V(j) + qi*pot_ij
                     Vi = vi + q_elec(j)*pot_ij
                  end if
               end do
               V(i) = V(i) + vi
            end do
         end do
      end if
   end do
end subroutine Qelec2Qelec_potential_@PBC@

!================================================================================
! Compute the Coulomb forces felt by each melt ions due to other melt ions
!DEC$ ATTRIBUTES NOINLINE :: melt_forces_@PBC@
subroutine melt_forces_@PBC@(localwork, ewald, box, ions, xyz_ions, &
      electrodes, xyz_atoms, q_elec, force,stress_tensor, compute_force)
   implicit none

   ! Parameters in
   ! -------------
   type(MW_localwork_t), intent(in)    :: localwork       !< Local work assignment
   type(MW_ewald_t),     intent(in)    :: ewald          !< Ewald summation parameters
   type(MW_box_t),       intent(in)    :: box            !< Simulation box parameters
   type(MW_ion_t),       intent(in)    :: ions(:)        !< ions parameters
   real(wp),             intent(in)    :: xyz_ions(:,:)  !< ions xyz positions
   type(MW_electrode_t), intent(inout) :: electrodes(:)   !< electrode parameters
   real(wp),             intent(in)    :: xyz_atoms(:,:)  !< electrode atoms xyz positions
   real(wp),             intent(in)    :: q_elec(:)      !< Electrode atoms charge
   logical,              intent(in)    :: compute_force

   ! Parameters out
   ! --------------
   real(wp),             intent(inout) :: force(:,:) !< Coulomb force on ions
   real(wp),             intent(inout) :: stress_tensor(:,:) !< Coulomb force contrib. to stress 

   ! Local
   ! -----
   integer :: num_ion_types, num_elec_types
   integer :: i, j, itype, jtype
   real(wp) :: a, b, c
   real(wp) :: qi, qj !< ion charge
   real(wp) :: Q_j    !< electrode charge
   real(wp) :: erfc_alpha, erfc_eta, exp_alpha, exp_eta
   real(wp) :: fij, fix, fiy, fiz
   real(wp) :: drnorm, drnorm2, drnorm2rec, drnormrec, dx, dy, dz
   real(wp) :: alpha, alphasq, alphapi
   real(wp) :: eta, etasq, etapi,etai,etaj,etaij,etaijsq,etaijpi

   integer :: num_block_diag, num_block_full
   integer :: iblock, iblock_offset, istart, iend, jstart, jend
   integer :: count_n, offset_n, count_m, offset_m
   real(wp) :: fx_melt2ele, fy_melt2ele, fz_melt2ele, fx_ele2ele, fy_ele2ele, fz_ele2ele

   num_ion_types = size(ions,1)
   num_elec_types = size(electrodes,1)

   ! Precompute some factors
   ! -----------------------
   a = box%length(1)
   b = box%length(2)
   c = box%length(3)

   alpha = ewald%alpha
   alphasq = alpha*alpha
   alphapi = 2.0_wp*alpha/sqrtpi


   ! Short range contribution
   ! ------------------------
   do itype = 1, num_ion_types
      count_n = ions(itype)%count
      offset_n = ions(itype)%offset
      qi = ions(itype)%charge
      if (qi /= 0.0_wp) then
         do jtype = 1, itype - 1
            count_m = ions(jtype)%count
            offset_m = ions(jtype)%offset
            qj = ions(jtype)%charge
            if ((qj /= 0.0_wp) .and. ewald%sr_interaction_ion2ion(itype,jtype)) then
               ! Number of blocks with full ion2ion interactions
               num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,jtype)
               iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,jtype)
               do iblock = 1, num_block_full
                  call update_other_block_boundaries(iblock_offset+iblock, &
                        offset_n, count_n, offset_m, count_m, &
                        istart, iend, jstart, jend)
                  do i = istart, iend
                     fix = 0.0_wp
                     fiy = 0.0_wp
                     fiz = 0.0_wp
                     do j = jstart, jend
                        call minimum_image_displacement_@PBC@(a, b, c, &
                              xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                              xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                              dx, dy, dz, drnorm2)
                        if (drnorm2 < ewald%rcutsq) then

                           fij = force_melt2melt_ij(drnorm2, alpha, alphasq, alphapi, qi, qj)
                           fix = fix + fij*dx
                           fiy = fiy + fij*dy
                           fiz = fiz + fij*dz

                           force(j,1) = force(j,1) + fij*dx
                           force(j,2) = force(j,2) + fij*dy
                           force(j,3) = force(j,3) + fij*dz

			   stress_tensor(1,1)=stress_tensor(1,1)+dx*dx*fij
			   stress_tensor(2,2)=stress_tensor(2,2)+dy*dy*fij
			   stress_tensor(3,3)=stress_tensor(3,3)+dz*dz*fij
			   stress_tensor(1,2)=stress_tensor(1,2)+dx*dy*fij
			   stress_tensor(1,3)=stress_tensor(1,3)+dx*dz*fij
			   stress_tensor(2,3)=stress_tensor(2,3)+dy*dz*fij			   
                        end if
                     end do
                     force(i,1) = force(i,1) - fix
                     force(i,2) = force(i,2) - fiy
                     force(i,3) = force(i,3) - fiz
                  end do
               end do
            end if
         end do

         if (ewald%sr_interaction_ion2ion(itype,itype)) then
            ! Blocks on the diagonal compute only half of the pair interactions
            qj = ions(itype)%charge
            num_block_diag = localwork%pair_ion2ion_num_block_diag_local(itype)
            iblock_offset = localwork%pair_ion2ion_diag_iblock_offset(itype)

            do iblock = 1, num_block_diag
               call update_diag_block_boundaries(iblock_offset+iblock, &
                     offset_n, count_n, istart, iend)

               do i = istart, iend
                  fix = 0.0_wp
                  fiy = 0.0_wp
                  fiz = 0.0_wp
                  do j = istart, iend
                     call minimum_image_displacement_@PBC@(a, b, c, &
                           xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                           xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                           dx, dy, dz, drnorm2)
                     if ((drnorm2 > 0.0_wp) .and. (drnorm2 < ewald%rcutsq)) then
                        fij = force_melt2melt_ij(drnorm2, alpha, alphasq, alphapi, qi, qj)
                        fix = fix + fij*dx
                        fiy = fiy + fij*dy
                        fiz = fiz + fij*dz


			stress_tensor(1,1)=stress_tensor(1,1)+0.5_wp*dx*dx*fij
			stress_tensor(2,2)=stress_tensor(2,2)+0.5_wp*dy*dy*fij
			stress_tensor(3,3)=stress_tensor(3,3)+0.5_wp*dz*dz*fij
			stress_tensor(1,2)=stress_tensor(1,2)+0.5_wp*dx*dy*fij
			stress_tensor(1,3)=stress_tensor(1,3)+0.5_wp*dx*dz*fij
			stress_tensor(2,3)=stress_tensor(2,3)+0.5_wp*dy*dz*fij			   
                     end if
                  end do
                  force(i,1) = force(i,1) - fix
                  force(i,2) = force(i,2) - fiy
                  force(i,3) = force(i,3) - fiz
               end do
            end do

            ! Number of blocks with full ion2ion interactions
            qi = ions(itype)%charge
            qj = ions(itype)%charge
            num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,itype)
            iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,itype)
            do iblock = 1, num_block_full
               call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
                     istart, iend, jstart, jend)
               do i = istart, iend
                  fix = 0.0_wp
                  fiy = 0.0_wp
                  fiz = 0.0_wp
                  do j = jstart, jend
                     call minimum_image_displacement_@PBC@(a, b, c, &
                           xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                           xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                           dx, dy, dz, drnorm2)
                     if (drnorm2 < ewald%rcutsq) then
                        fij = force_melt2melt_ij(drnorm2, alpha, alphasq, alphapi, qi, qj)
                        fix = fix + fij*dx
                        fiy = fiy + fij*dy
                        fiz = fiz + fij*dz

                        force(j,1) = force(j,1) + fij*dx
                        force(j,2) = force(j,2) + fij*dy
                        force(j,3) = force(j,3) + fij*dz

			stress_tensor(1,1)=stress_tensor(1,1)+dx*dx*fij
			stress_tensor(2,2)=stress_tensor(2,2)+dy*dy*fij
			stress_tensor(3,3)=stress_tensor(3,3)+dz*dz*fij
			stress_tensor(1,2)=stress_tensor(1,2)+dx*dy*fij
			stress_tensor(1,3)=stress_tensor(1,3)+dx*dz*fij
			stress_tensor(2,3)=stress_tensor(2,3)+dy*dz*fij			   
                     end if
                  end do
                  force(i,1) = force(i,1) - fix
                  force(i,2) = force(i,2) - fiy
                  force(i,3) = force(i,3) - fiz
               end do
            end do
         end if

         qi = ions(itype)%charge
         do jtype = 1, num_elec_types
            if (ewald%sr_interaction_atom2ion(itype,jtype)) then
               count_m = electrodes(jtype)%count
               offset_m = electrodes(jtype)%offset
               ! Number of blocks with full ion2ion interactions
               num_block_full = localwork%pair_atom2ion_num_block_full_local(itype,jtype)
               iblock_offset = localwork%pair_atom2ion_full_iblock_offset(itype,jtype)
               eta = electrodes(jtype)%eta
               etasq = eta*eta
               etapi = 2.0*eta / sqrtpi
               fx_melt2ele = 0.0_wp
               fy_melt2ele = 0.0_wp
               fz_melt2ele = 0.0_wp
               do iblock = 1, num_block_full
                  call update_other_block_boundaries(iblock_offset+iblock, &
                        offset_n, count_n, offset_m, count_m, &
                        istart, iend, jstart, jend)
                  do i = istart, iend
                     fix = 0.0_wp
                     fiy = 0.0_wp
                     fiz = 0.0_wp
                     do j = jstart, jend
                        call minimum_image_displacement_@PBC@(a, b, c, &
                              xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                              xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                              dx, dy, dz, drnorm2)
                        if (drnorm2 < ewald%rcutsq) then
                           Q_j = q_elec(j)
                           fij = force_elec2melt_ij(drnorm2, alpha, alphasq, alphapi, eta, etasq, etapi, qi, Q_j)
                           fix = fix + fij*dx
                           fiy = fiy + fij*dy
                           fiz = fiz + fij*dz

			   stress_tensor(1,1)=stress_tensor(1,1)+dx*dx*fij
			   stress_tensor(2,2)=stress_tensor(2,2)+dy*dy*fij
			   stress_tensor(3,3)=stress_tensor(3,3)+dz*dz*fij
			   stress_tensor(1,2)=stress_tensor(1,2)+dx*dy*fij
			   stress_tensor(1,3)=stress_tensor(1,3)+dx*dz*fij
			   stress_tensor(2,3)=stress_tensor(2,3)+dy*dz*fij			   
                        end if
                     end do
                     force(i,1) = force(i,1) - fix
                     force(i,2) = force(i,2) - fiy
                     force(i,3) = force(i,3) - fiz
                     fx_melt2ele = fx_melt2ele + fix
                     fy_melt2ele = fy_melt2ele + fiy
                     fz_melt2ele = fz_melt2ele + fiz
                  end do
                  !acc end parallel loop
               end do
               electrodes(jtype)%force_ions(1,3) = electrodes(jtype)%force_ions(1,3) + fx_melt2ele
               electrodes(jtype)%force_ions(2,3) = electrodes(jtype)%force_ions(2,3) + fy_melt2ele
               electrodes(jtype)%force_ions(3,3) = electrodes(jtype)%force_ions(3,3) + fz_melt2ele
            end if
         end do
      end if
   end do

   ! Elec -> Elec interactions
   if (compute_force) then
      do itype = 1, num_elec_types
         count_n = electrodes(itype)%count
         offset_n = electrodes(itype)%offset
         do jtype = 1, itype-1
            count_m =  electrodes(jtype)%count
            offset_m = electrodes(jtype)%offset
            if (ewald%sr_interaction_atom2atom(itype,jtype)) then
               etai = electrodes(itype)%eta
               etaj = electrodes(jtype)%eta
               etaij = (etai*etaj) / sqrt(etai*etai + etaj*etaj)
               etaijsq = etaij * etaij
               etaijpi = 2.0*etaij / sqrtpi
 
               ! Number of blocks with full interactions (below the diagonal
               num_block_full = localwork%pair_atom2atom_num_block_full_local(itype,jtype)
               iblock_offset = localwork%pair_atom2atom_full_iblock_offset(itype,jtype)
 
               do iblock = 1, num_block_full
                  call update_other_block_boundaries(iblock_offset+iblock, &
                        offset_n, count_n, offset_m, count_m, istart, iend, jstart, jend)
                  do i = istart, iend
                     fx_ele2ele = 0.0_wp
                     fy_ele2ele = 0.0_wp
                     fz_ele2ele = 0.0_wp
                     do j = jstart, jend
                        call minimum_image_displacement_@PBC@(a, b, c, &
                              xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                              xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                              dx, dy, dz, drnorm2)
                        if (drnorm2 < ewald%rcutsq)  then
                           drnorm = sqrt(drnorm2)
                           drnormrec = 1.0_wp / drnorm
                           drnorm2rec = drnormrec * drnormrec
                           ! Change erfc to 1-erf because of underflows
                           erfc_alpha = erfc(alpha*drnorm)*drnormrec
                           exp_alpha = exp(-alphasq*drnorm2)*alphapi
                           erfc_eta = erfc(etaij*drnorm)*drnormrec
                           exp_eta = exp(-etaijsq*drnorm2)*etaijpi
                           fij = q_elec(i) * q_elec(j) * drnorm2rec * ((erfc_alpha+exp_alpha) - (erfc_eta+exp_eta))
                           electrodes(jtype)%force_ions(1,2) = electrodes(jtype)%force_ions(1,2) + fij * dx
                           electrodes(jtype)%force_ions(2,2) = electrodes(jtype)%force_ions(2,2) + fij * dy
                           electrodes(jtype)%force_ions(3,2) = electrodes(jtype)%force_ions(3,2) + fij * dz
                           fx_ele2ele = fx_ele2ele + fij * dx
                           fy_ele2ele = fy_ele2ele + fij * dy
                           fz_ele2ele = fz_ele2ele + fij * dz
                        end if
                     end do
                     electrodes(itype)%force_ions(1,2) = electrodes(itype)%force_ions(1,2) - fx_ele2ele
                     electrodes(itype)%force_ions(2,2) = electrodes(itype)%force_ions(2,2) - fy_ele2ele
                     electrodes(itype)%force_ions(3,2) = electrodes(itype)%force_ions(3,2) - fz_ele2ele
                  end do
               end do
            end if
         end do
 
         if (ewald%sr_interaction_atom2atom(itype,itype)) then
 
            etai = electrodes(itype)%eta
            etaij = etai / sqrt(2.0_wp)
            etaijsq = etaij * etaij
            etaijpi = 2.0*etaij / sqrtpi
 
            ! Blocks on the diagonal compute only half of the pair interactions
            num_block_diag = localwork%pair_atom2atom_num_block_diag_local(itype)
            iblock_offset = localwork%pair_atom2atom_diag_iblock_offset(itype)
 
            do iblock = 1, num_block_diag
               call update_diag_block_boundaries(iblock+iblock_offset, &
                     offset_n, count_n, istart, iend)
               do i = istart, iend
                  fx_ele2ele = 0.0_wp
                  fy_ele2ele = 0.0_wp
                  fz_ele2ele = 0.0_wp
                  do j = istart, iend
                     call minimum_image_displacement_@PBC@(a, b, c, &
                           xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                           xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                           dx, dy, dz, drnorm2)
                     if ((drnorm2 > 0.0_wp) .and. (drnorm2 < ewald%rcutsq))  then
                        drnorm = sqrt(drnorm2)
                        drnormrec = 1.0_wp / drnorm
                        drnorm2rec = drnormrec * drnormrec
                        ! Change erfc to 1-erf because of underflows
                        erfc_alpha = erfc(alpha*drnorm)*drnormrec
                        exp_alpha = exp(-alphasq*drnorm2)*alphapi
                        erfc_eta = erfc(etaij*drnorm)*drnormrec
                        exp_eta = exp(-etaijsq*drnorm2)*etaijpi
                        fij = q_elec(i) * q_elec(j) * drnorm2rec * ((erfc_alpha+exp_alpha) - (erfc_eta+exp_eta))
                        fx_ele2ele = fx_ele2ele + fij * dx
                        fy_ele2ele = fy_ele2ele + fij * dy
                        fz_ele2ele = fz_ele2ele + fij * dz
                     end if
                  end do
                  electrodes(itype)%force_ions(1,2) = electrodes(itype)%force_ions(1,2) - fx_ele2ele
                  electrodes(itype)%force_ions(2,2) = electrodes(itype)%force_ions(2,2) - fy_ele2ele
                  electrodes(itype)%force_ions(3,2) = electrodes(itype)%force_ions(3,2) - fz_ele2ele
               end do
            end do
 
            ! Number of blocks with full interactions (below the diagonal
            num_block_full = localwork%pair_atom2atom_num_block_full_local(itype,itype)
            iblock_offset = localwork%pair_atom2atom_full_iblock_offset(itype,itype)
 
            do iblock = 1, num_block_full
               call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
                     istart, iend, jstart, jend)
               do i = istart, iend
                  fx_ele2ele = 0.0_wp
                  fy_ele2ele = 0.0_wp
                  fz_ele2ele = 0.0_wp
                  do j = jstart, jend
                     call minimum_image_displacement_@PBC@(a, b, c, &
                           xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                           xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                           dx, dy, dz, drnorm2)
                     if (drnorm2 < ewald%rcutsq)  then
                        drnorm = sqrt(drnorm2)
                        drnormrec = 1.0_wp / drnorm
                        drnorm2rec = drnormrec * drnormrec
                        ! Change erfc to 1-erf because of underflows
                        erfc_alpha = erfc(alpha*drnorm)*drnormrec
                        exp_alpha = exp(-alphasq*drnorm2)*alphapi
                        erfc_eta = erfc(etaij*drnorm)*drnormrec
                        exp_eta = exp(-etaijsq*drnorm2)*etaijpi
                        fij = q_elec(i) * q_elec(j) * drnorm2rec * ((erfc_alpha+exp_alpha) - (erfc_eta+exp_eta))
                        electrodes(itype)%force_ions(1,2) = electrodes(itype)%force_ions(1,2) + fij * dx
                        electrodes(itype)%force_ions(2,2) = electrodes(itype)%force_ions(2,2) + fij * dy
                        electrodes(itype)%force_ions(3,2) = electrodes(itype)%force_ions(3,2) + fij * dz
                        fx_ele2ele = fx_ele2ele + fij * dx
                        fy_ele2ele = fy_ele2ele + fij * dy
                        fz_ele2ele = fz_ele2ele + fij * dz
                     end if
                  end do
                  electrodes(itype)%force_ions(1,2) = electrodes(itype)%force_ions(1,2) - fx_ele2ele
                  electrodes(itype)%force_ions(2,2) = electrodes(itype)%force_ions(2,2) - fy_ele2ele
                  electrodes(itype)%force_ions(3,2) = electrodes(itype)%force_ions(3,2) - fz_ele2ele
               end do
            end do
         end if
      end do
   end if

end subroutine melt_forces_@PBC@

!================================================================================
! Compute the Energy due to Coulomb interactions
!DEC$ ATTRIBUTES NOINLINE :: energy_@PBC@
subroutine energy_@PBC@(localwork, ewald, box, ions, xyz_ions, &
      electrodes, xyz_atoms, q_elec, h)
   implicit none

   ! Parameters in
   ! -------------
   type(MW_localwork_t), intent(in)    :: localwork       !< Local work assignment
   type(MW_ewald_t),     intent(in)    :: ewald          !< Ewald summation parameters
   type(MW_box_t),       intent(in)    :: box            !< Simulation box parameters
   type(MW_ion_t),       intent(in)    :: ions(:)        !< ions parameters
   real(wp),             intent(in)    :: xyz_ions(:,:)  !< ions xyz positions
   type(MW_electrode_t), intent(in)    :: electrodes(:)   !< electrode parameters
   real(wp),             intent(in)    :: xyz_atoms(:,:)  !< electrode atoms xyz positions
   real(wp),             intent(in)    :: q_elec(:)      !< Electrode atoms charge

   ! Parameters out
   ! --------------
   real(wp),             intent(inout) :: h !< Coulomb energy

   ! Local
   ! -----
   integer :: num_ion_types, num_elec_types
   integer :: i, j, itype, jtype
   real(wp) :: a, b, c
   real(wp) :: qi, qj !< ion charge
   real(wp) :: erfc_ij, erfdiff
   real(wp) :: vi, pot_ij
   real(wp) :: drnorm, drnorm2, drnormrec, rcutsq
   real(wp) :: alpha, alphasq, alphapi
   real(wp) :: etai, etaj, etaij, etaijsq
   integer :: num_block_diag, num_block_full
   integer :: iblock, iblock_offset, istart, iend, jstart, jend
   integer :: count_n, offset_n, count_m, offset_m

   num_ion_types = size(ions,1)
   num_elec_types = size(electrodes,1)

   ! Precompute some factors
   ! -----------------------
   a = box%length(1)
   b = box%length(2)
   c = box%length(3)

   alpha = ewald%alpha
   alphasq = alpha*alpha
   alphapi = 2.0_wp*alpha/sqrtpi

   rcutsq = ewald%rcutsq

   ! Zero energy
   ! ----------------
   h = 0.0_wp
   vi = 0.0_wp


   ! Short range contribution from Melt->Melt and Elec->Melt interaction
   ! -------------------------------------------------------------------
   do itype = 1, num_ion_types
      count_n = ions(itype)%count
      offset_n = ions(itype)%offset
      qi = ions(itype)%charge
      if (qi /= 0.0_wp) then
         do jtype = 1, itype - 1
            count_m = ions(jtype)%count
            offset_m = ions(jtype)%offset
            qj = ions(jtype)%charge
            if ((qj /= 0.0_wp) .and. ewald%sr_interaction_ion2ion(itype,jtype)) then
               ! Number of blocks with full ion2ion interactions
               num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,jtype)
               iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,jtype)

               do iblock = 1, num_block_full
                  call update_other_block_boundaries(iblock+iblock_offset, &
                        offset_n, count_n, offset_m, count_m, &
                        istart, iend, jstart, jend)
                  do i = istart, iend
                     do j = jstart, jend
                        call minimum_image_distance_@PBC@(a, b, c, &
                              xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                              xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                              drnorm2)
                        if (drnorm2 < rcutsq) then
                           drnorm = sqrt(drnorm2)
                           drnormrec = 1.0_wp/drnorm
                           erfc_ij = erfc(alpha*drnorm)
                           vi = vi + qi*qj*erfc_ij*drnormrec
                        end if
                     end do
                  end do
               end do
            end if
         end do

         if (ewald%sr_interaction_ion2ion(itype,itype)) then
            ! Blocks on the diagonal compute only half of the pair interactions
            qj = ions(itype)%charge
            num_block_diag = localwork%pair_ion2ion_num_block_diag_local(itype)
            iblock_offset = localwork%pair_ion2ion_diag_iblock_offset(itype)
            do iblock = 1, num_block_diag
               call update_diag_block_boundaries(iblock_offset+iblock, offset_n, count_n, istart, iend)
               do i = istart, iend
                  do j = istart, iend
                     call minimum_image_distance_@PBC@(a, b, c, &
                           xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                           xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                           drnorm2)
                     if ((drnorm2 > 0.0_wp) .and. (drnorm2 < rcutsq)) then
                        drnorm = sqrt(drnorm2)
                        drnormrec = 1.0_wp/drnorm
                        erfc_ij = erfc(alpha*drnorm)
                        vi = vi + 0.5_wp*qi*qj*erfc_ij*drnormrec
                     end if
                  end do
               end do
            end do

            qi = ions(itype)%charge
            qj = ions(itype)%charge
            ! Number of blocks with full ion2ion interactions
            num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,itype)
            iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,itype)

            do iblock = 1, num_block_full
               call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
                     istart, iend, jstart, jend)
               do i = istart, iend
                  do j = jstart, jend
                     call minimum_image_distance_@PBC@(a, b, c, &
                           xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                           xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                           drnorm2)
                     if (drnorm2 < rcutsq) then
                        drnorm = sqrt(drnorm2)
                        drnormrec = 1.0_wp/drnorm
                        erfc_ij = erfc(alpha*drnorm)
                        vi = vi + qi*qj*erfc_ij*drnormrec
                     end if
                  end do
               end do
            end do
         end if

         qi = ions(itype)%charge
         do jtype = 1, num_elec_types
            count_m = electrodes(jtype)%count
            offset_m = electrodes(jtype)%offset
            etaj = electrodes(jtype)%eta
            if (ewald%sr_interaction_atom2ion(itype,jtype)) then
               ! Number of blocks with full ion2ion interactions
               num_block_full = localwork%pair_atom2ion_num_block_full_local(itype,jtype)
               iblock_offset = localwork%pair_atom2ion_full_iblock_offset(itype,jtype)

               do iblock = 1, num_block_full
                  call update_other_block_boundaries(iblock+iblock_offset, &
                        offset_n, count_n, offset_m, count_m, &
                        istart, iend, jstart, jend)
                  do i = istart, iend
                     do j = jstart, jend
                        call minimum_image_distance_@PBC@(a, b, c, &
                              xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                              xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                              drnorm2)
                        if (drnorm2 < rcutsq) then
                           drnorm = sqrt(drnorm2)
                           drnormrec = 1.0_wp/drnorm
                           ! Change erfc to 1-erf because of underflows
                           erfdiff = (erfc(alpha*drnorm) - erfc(etaj*drnorm))
                           vi = vi + qi*q_elec(j)*erfdiff*drnormrec
                        end if
                     end do
                  end do
               end do
            end if
         end do
      end if
   end do

   ! Elec -> Elec interactions
   do itype = 1, num_elec_types
      count_n = electrodes(itype)%count
      offset_n = electrodes(itype)%offset
      do jtype = 1, itype-1
         count_m =  electrodes(jtype)%count
         offset_m = electrodes(jtype)%offset
         if (ewald%sr_interaction_atom2atom(itype,jtype)) then
            etai = electrodes(itype)%eta
            etaj = electrodes(jtype)%eta
            etaij = (etai*etaj)/sqrt(etai*etai + etaj*etaj)
            etaijsq = etaij*etaij

            ! Number of blocks with full interactions (below the diagonal
            num_block_full = localwork%pair_atom2atom_num_block_full_local(itype,jtype)
            iblock_offset = localwork%pair_atom2atom_full_iblock_offset(itype,jtype)

            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
                     offset_n, count_n, offset_m, count_m, istart, iend, jstart, jend)
               do i = istart, iend
                  do j = jstart, jend
                     call minimum_image_distance_@PBC@(a, b, c, &
                           xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                           xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                           drnorm2)
                     if (drnorm2 < rcutsq)  then
                        drnorm = sqrt(drnorm2)
                        ! Change erfc to 1-erf because of underflows
                        ! erfdiff = exp(-alphasq*drnorm2) * &
                        !       (erfc_scaled(alpha*drnorm) - exp((alphasq - etaijsq)*drnorm2) * erfc_scaled(etaij*drnorm))
                        ! erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                        erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                        pot_ij =  erfdiff/drnorm
                        Vi = Vi + q_elec(i)*q_elec(j)*pot_ij
                     end if
                  end do
               end do
            end do
         end if
      end do

      if (ewald%sr_interaction_atom2atom(itype,itype)) then

         etai = electrodes(itype)%eta
         etaij = etai/sqrt(2.0_wp)
         etaijsq = etaij*etaij

         ! Blocks on the diagonal compute only half of the pair interactions
         num_block_diag = localwork%pair_atom2atom_num_block_diag_local(itype)
         iblock_offset = localwork%pair_atom2atom_diag_iblock_offset(itype)

         do iblock = 1, num_block_diag
            call update_diag_block_boundaries(iblock+iblock_offset, &
                  offset_n, count_n, istart, iend)
            do i = istart, iend
               do j = istart, iend
                  call minimum_image_distance_@PBC@(a, b, c, &
                        xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                        xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                        drnorm2)
                  if ((drnorm2 > 0.0_wp) .and. (drnorm2 < rcutsq))  then
                     drnorm = sqrt(drnorm2)
                     erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                     pot_ij =  erfdiff/drnorm
                     vi = vi + 0.5_wp*q_elec(i)*q_elec(j)*pot_ij
                  end if
               end do
            end do
         end do

         ! Number of blocks with full interactions (below the diagonal
         num_block_full = localwork%pair_atom2atom_num_block_full_local(itype,itype)
         iblock_offset = localwork%pair_atom2atom_full_iblock_offset(itype,itype)

         do iblock = 1, num_block_full
            call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
                  istart, iend, jstart, jend)
            do i = istart, iend
               do j = jstart, jend
                  call minimum_image_distance_@PBC@(a, b, c, &
                        xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                        xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                        drnorm2)
                  if (drnorm2 < rcutsq)  then
                     drnorm = sqrt(drnorm2)
                     erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                     pot_ij =  erfdiff/drnorm
                     vi = vi + q_elec(i)*q_elec(j)*pot_ij
                  end if
               end do
            end do
         end do
      end if
   end do
   h = vi
end subroutine energy_@PBC@

! ==========================================================================================
! Determine if short-range coulomb interaction should be taken into account between 2 particle types
!
! By default short-range coulomb interaction is taken into account
! For fixed particles, short-range coulomb interaction is not taken
! into account between 2 particle types if the minimum distance
! between 2 particles of the distinct sets is greater than the
! specified cut-off (rcut)
subroutine setup_sr_interaction_@PBC@(ewald, ions, electrodes, &
      xyz_ions, xyz_atoms, box)
   implicit none
   ! Parameters
   ! ----------
   type(MW_ewald_t), intent(inout) :: ewald           !< Coulomb ewald structure
   type(MW_ion_t), intent(in) :: ions(:)             !< ions parameters
   type(MW_electrode_t), intent(in) :: electrodes(:) !< electrode parameters
   real(wp), intent(in) :: xyz_ions(:,:)             !< ions xyz coordinates
   real(wp), intent(in) :: xyz_atoms(:,:)            !< atoms xyz coordinates
   type(MW_box_t), intent(in) :: box                 !< box parameters

   ! Local
   ! -----
   integer :: num_ion_types, num_atom_types
   integer :: ios, itype, jtype, i, j
   real(wp) :: a, b, c
   real(wp) :: drnorm2

   a = box%length(1)
   b = box%length(2)
   c = box%length(3)

   num_ion_types = size(ions,1)
   num_atom_types = size(electrodes,1)
   allocate(ewald%sr_interaction_ion2ion(num_ion_types, num_ion_types), &
         ewald%sr_interaction_atom2atom(num_atom_types, num_atom_types), &
         ewald%sr_interaction_ion2atom(num_atom_types, num_ion_types), &
         ewald%sr_interaction_atom2ion(num_ion_types, num_atom_types), stat=ios)
   if (ios /= 0) then
      call MW_errors_allocate_error("setup_sr_interaction", "ewald.f90", ios)
   end if
   ewald%sr_interaction_ion2ion(:,:) = .true.
   ewald%sr_interaction_atom2atom(:,:) = .true.
   ewald%sr_interaction_ion2atom(:,:) = .true.
   ewald%sr_interaction_atom2ion(:,:) = .true.

   ! Check fixed ion minimum distance
   do itype = 1, num_ion_types
      if (.not. ions(itype)%mobile) then
         do jtype = 1, itype-1
            if (.not. ions(jtype)%mobile) then
               ! Assume no interaction and check if there is interaction
               ewald%sr_interaction_ion2ion(itype,jtype) = .false.
               ewald%sr_interaction_ion2ion(jtype,itype) = .false.
               ! Look for 1st interacting pair and reset interaction to .true. if we find one
               Ion2IonMinimumDistanceLoop: do i = ions(itype)%offset + 1, ions(itype)%offset + ions(itype)%count
                  do j = ions(jtype)%offset + 1, ions(jtype)%offset + ions(jtype)%count
                     call minimum_image_distance_@PBC@(a, b, c, &
                           xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                           xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                           drnorm2)
                     if (drnorm2 < ewald%rcutsq) then
                        ewald%sr_interaction_ion2ion(itype,jtype) = .true.
                        ewald%sr_interaction_ion2ion(jtype,itype) = .true.
                        exit Ion2IonMinimumDistanceLoop
                     end if
                  end do
               end do Ion2IonMinimumDistanceLoop
            end if
         end do

         ! Assume no self-interaction and check if there is interaction
         ewald%sr_interaction_ion2ion(itype,itype) = .false.
         ! Look for 1st interacting pair and reset interaction to .true. if we find one
         SelfIonMinimumDistanceLoop: do i = ions(itype)%offset+1, ions(itype)%offset + ions(itype)%count
            do j = ions(itype)%offset+1, i-1
               call minimum_image_distance_@PBC@(a, b, c, &
                     xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                     xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), drnorm2)
               if (drnorm2 < ewald%rcutsq) then
                  ewald%sr_interaction_ion2ion(itype,itype) = .true.
                  exit SelfIonMinimumDistanceLoop
               end if
            end do
         end do SelfIonMinimumDistanceLoop

         do jtype = 1, num_atom_types
            ! Assume no interaction and check if there is interaction
            ewald%sr_interaction_atom2ion(itype,jtype) = .false.
            ewald%sr_interaction_ion2atom(jtype,itype) = .false.
            ! Look for 1st interacting pair and reset interaction to .true. if we find one
            Ion2AtomMinimumDistanceLoop: do i = ions(itype)%offset+1, ions(itype)%offset + ions(itype)%count
               do j = electrodes(jtype)%offset+1, electrodes(jtype)%offset+ electrodes(jtype)%count
                  call minimum_image_distance_@PBC@(a, b, c, &
                        xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                        xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), drnorm2)
                  if (drnorm2 < ewald%rcutsq) then
                     ewald%sr_interaction_atom2ion(itype,jtype) = .true.
                     ewald%sr_interaction_ion2atom(jtype,itype) = .true.
                     exit Ion2AtomMinimumDistanceLoop
                  end if
               end do
            end do Ion2AtomMinimumDistanceLoop
         end do
      end if
   end do

   do itype = 1, num_atom_types
      do jtype = 1, itype-1
         ! Assume no interaction and check if there is interaction
         ewald%sr_interaction_atom2atom(itype,jtype) = .false.
         ewald%sr_interaction_atom2atom(jtype,itype) = .false.
         ! Look for 1st interacting pair and reset interaction to .true. if we find one
         Atom2AtomMinimumDistanceLoop: do i = electrodes(itype)%offset+1, electrodes(itype)%offset + electrodes(itype)%count
            do j = electrodes(jtype)%offset+1, electrodes(jtype)%offset + electrodes(jtype)%count
               call minimum_image_distance_@PBC@(a, b, c, &
                     xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                     xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), drnorm2)
               if (drnorm2 < ewald%rcutsq) then
                  ewald%sr_interaction_atom2atom(itype,jtype) = .true.
                  ewald%sr_interaction_atom2atom(jtype,itype) = .true.
                  exit Atom2AtomMinimumDistanceLoop
               end if
            end do
         end do Atom2AtomMinimumDistanceLoop
      end do

      ! Check for self-interaction
      ! Assume no interaction and check if there is interaction
      ewald%sr_interaction_atom2atom(itype,jtype) = .false.
      ! Look for 1st interacting pair and reset interaction to .true. if we find one
      SelfAtomMinimumDistanceLoop: do i = electrodes(itype)%offset+1, electrodes(itype)%offset+electrodes(itype)%count
         do j = electrodes(itype)%offset+1, i-1
            call minimum_image_distance_@PBC@(a, b, c, &
                  xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                  xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), drnorm2)
            if (drnorm2 < ewald%rcutsq) then
               ewald%sr_interaction_atom2atom(itype,jtype) = .true.
               exit SelfAtomMinimumDistanceLoop
            end if
         end do
      end do SelfAtomMinimumDistanceLoop
   end do
end subroutine setup_sr_interaction_@PBC@
