subroutine gradQelec_electricfield_@PBC@(localwork, ewald, box, ions, xyz_ions, &
     electrodes, xyz_atoms, gradQ_efield)
   implicit none

   ! Parameters in
   ! -------------
   type(MW_localwork_t), intent(in)    :: localwork       !< Local work assignment
   type(MW_ewald_t),     intent(in)    :: ewald          !< Ewald summation parameters
   type(MW_box_t),       intent(in)    :: box            !< Simulation box parameters
   type(MW_ion_t),       intent(in)    :: ions(:)        !< ions parameters
   real(wp),             intent(in)    :: xyz_ions(:,:)  !< ions xyz positions
   type(MW_electrode_t), intent(in)    :: electrodes(:)   !< electrode parameters
   real(wp),             intent(in)    :: xyz_atoms(:,:)  !< electrode atoms xyz positions

   ! Parameters out
   ! --------------
   real(wp),             intent(inout) :: gradQ_efield(:,:) !< Electric field on ions

   ! Local
   ! -----
   integer :: num_ions, num_ion_types, num_elec_types
   integer :: i, j, jx, jy, jz, itype, jtype
   real(wp) :: a, b, c
   real(wp) :: qj !< ion charge
   real(wp) :: dx, dy, dz
   real(wp) :: drnorm, drnorm2
   real(wp) :: drnormrec, drnorm2rec
   real(wp) :: alpha, alphasq, alphapi
   real(wp) :: eta, etasq, etapi
   real(wp) :: erfc_alpha, erfc_eta, exp_alpha, exp_eta
   real(wp) :: eij

   integer :: num_block_full
   integer :: iblock, iblock_offset, istart, iend, jstart, jend
   integer :: count_n, offset_n, count_m, offset_m

   num_ion_types = size(ions,1)
   num_elec_types = size(electrodes,1)
   num_ions = size(gradQ_efield,1)/3

   ! Precompute some factors
   ! -----------------------
   a = box%length(1)
   b = box%length(2)
   c = box%length(3)

   alpha = ewald%alpha
   alphasq = alpha*alpha
   alphapi = 2.0_wp*alpha/sqrtpi


   ! Short range contribution
   ! ------------------------
   do itype = 1, num_elec_types
      count_n = electrodes(itype)%count
      offset_n = electrodes(itype)%offset
      eta = electrodes(itype)%eta
      etasq = eta*eta
      etapi = 2.0_wp*eta/sqrtpi
      do jtype = 1, num_ion_types
         qj = ions(jtype)%charge
         if ((qj /= 0.0_wp) .and. ewald%sr_interaction_ion2atom(itype,jtype)) then
            count_m = ions(jtype)%count
            offset_m = ions(jtype)%offset
            ! Number of blocks with full ion2ion interactions
            num_block_full = localwork%pair_ion2atom_num_block_full_local(itype,jtype)
            iblock_offset = localwork%pair_ion2atom_full_iblock_offset(itype,jtype)
            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
                    offset_n, count_n, offset_m, count_m, &
                    istart, iend, jstart, jend)
               do i = istart, iend
                  do j = jstart, jend
                     jx = j
                     jy = j + num_ions
                     jz = j + num_ions*2
                     call minimum_image_displacement_@PBC@(a, b, c, &
                          xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                          xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                          dx, dy, dz, drnorm2)
                     if (drnorm2 < ewald%rcutsq) then
                        drnorm = sqrt(drnorm2)
                        drnormrec = 1.0_wp/drnorm
                        drnorm2rec = drnormrec*drnormrec
                        ! Change erfc to 1-erf because of underflows
                        erfc_alpha = erfc(alpha*drnorm)*drnormrec
                        exp_alpha = exp(-alphasq*drnorm2)*alphapi
                        erfc_eta = erfc(eta*drnorm)*drnormrec
                        exp_eta = exp(-etasq*drnorm2)*etapi
                        eij =  drnorm2rec*((erfc_alpha + exp_alpha) - (erfc_eta + exp_eta))

                        gradQ_efield(jx,i) = gradQ_efield(jx,i) + eij*dx
                        gradQ_efield(jy,i) = gradQ_efield(jy,i) + eij*dy
                        gradQ_efield(jz,i) = gradQ_efield(jz,i) + eij*dz
                     end if
                  end do
               end do
               !acc end loop
            end do
         end if
      end do
   end do

end subroutine gradQelec_electricfield_@PBC@

!================================================================================
! Compute the gradient of the electric field for each ion in the melt with respect
! to variations of other dipoles in the melt
subroutine gradmumelt_electricfield_@PBC@(localwork, ewald, box, ions, xyz_ions, &
     gradmu_efield)
  implicit none

  ! Parameters in
  ! -------------
  type(MW_localwork_t),  intent(in)    :: localwork      !< Local work assignment
  type(MW_ewald_t),      intent(in)    :: ewald          !< Ewald summation parameters
  type(MW_box_t),        intent(in)    :: box            !< Simulation box parameters
  type(MW_ion_t),        intent(in)    :: ions(:)        !< ions parameters
  real(wp),              intent(in)    :: xyz_ions(:,:)  !< ions xyz positions

  ! Parameters out
  ! --------------
  real(wp),              intent(inout) :: gradmu_efield(:,:) !< matrix containing the
  ! gradient of the field

  ! Local
  ! -----
  integer :: num_ion_types, num_ions
  integer :: i, j, itype, jtype, ix, iy, iz, jx, jy, jz
  real(wp) :: a, b, c
  real(wp) :: drnorm2, dx, dy, dz
  real(wp) :: alpha, alphasq, alphapi
  real(wp), dimension(6) :: gmueij

  integer :: num_block_diag, num_block_full
  integer :: iblock, iblock_offset, istart, iend, jstart, jend
  integer :: count_n, offset_n, count_m, offset_m

  num_ion_types = size(ions,1)
  num_ions = size(gradmu_efield,1)/3

  ! Precompute some factors
  ! -----------------------
  a = box%length(1)
  b = box%length(2)
  c = box%length(3)

  alpha = ewald%alpha
  alphasq = alpha*alpha
  alphapi = 2.0_wp*alpha/sqrtpi


  ! Short range contribution
  ! ------------------------
  do itype = 1, num_ion_types
     count_n = ions(itype)%count
     offset_n = ions(itype)%offset
     do jtype = 1, itype - 1
        count_m = ions(jtype)%count
        offset_m = ions(jtype)%offset
        if (ewald%sr_interaction_ion2ion(itype,jtype)) then
           ! Number of blocks with full ion2ion interactions
           num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,jtype)
           iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,jtype)
           do iblock = 1, num_block_full
              call update_other_block_boundaries(iblock_offset+iblock, &
                   offset_n, count_n, offset_m, count_m, &
                   istart, iend, jstart, jend)
              do i = istart, iend
                 ix = i
                 iy = i + num_ions
                 iz = i + num_ions*2
                 do j = jstart, jend
                    jx = j
                    jy = j + num_ions
                    jz = j + num_ions*2
                    call minimum_image_displacement_@PBC@(a, b, c, &
                         xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                         xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                         dx, dy, dz, drnorm2)
                    if (drnorm2 < ewald%rcutsq) then
                       gmueij = gradmumelt_elec_ij(drnorm2, dx, dy, dz, &
                            alpha, alphasq, alphapi)
                       gradmu_efield(jx,ix) = gradmu_efield(jx,ix) + gmueij(1)
                       gradmu_efield(jy,ix) = gradmu_efield(jy,ix) + gmueij(2)
                       gradmu_efield(jz,ix) = gradmu_efield(jz,ix) + gmueij(3)

                       gradmu_efield(jy,iy) = gradmu_efield(jy,iy) + gmueij(4)
                       gradmu_efield(jz,iy) = gradmu_efield(jz,iy) + gmueij(5)

                       gradmu_efield(jz,iz) = gradmu_efield(jz,iz) + gmueij(6)
                    end if
                 end do
              end do
           end do
        end if
     end do

     if (ewald%sr_interaction_ion2ion(itype,itype)) then
        ! Blocks on the diagonal compute only half of the pair interactions
        num_block_diag = localwork%pair_ion2ion_num_block_diag_local(itype)
        iblock_offset = localwork%pair_ion2ion_diag_iblock_offset(itype)

        do iblock = 1, num_block_diag
           call update_diag_block_boundaries(iblock_offset+iblock, &
                offset_n, count_n, istart, iend)
           do i = istart, iend
              ix = i
              iy = i + num_ions
              iz = i + num_ions*2
              do j = istart, iend
                 jx = j
                 jy = j + num_ions
                 jz = j + num_ions*2
                 call minimum_image_displacement_@PBC@(a, b, c, &
                      xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                      xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                      dx, dy, dz, drnorm2)
                 if ((drnorm2 > 0.0_wp) .and. (drnorm2 < ewald%rcutsq)) then
                    gmueij = gradmumelt_elec_ij(drnorm2, dx, dy, dz, &
                         alpha, alphasq, alphapi)
                    gradmu_efield(jx,ix) = gradmu_efield(jx,ix) + gmueij(1)
                    gradmu_efield(jy,ix) = gradmu_efield(jy,ix) + gmueij(2)
                    gradmu_efield(jz,ix) = gradmu_efield(jz,ix) + gmueij(3)

                    gradmu_efield(jy,iy) = gradmu_efield(jy,iy) + gmueij(4)
                    gradmu_efield(jz,iy) = gradmu_efield(jz,iy) + gmueij(5)

                    gradmu_efield(jz,iz) = gradmu_efield(jz,iz) + gmueij(6)
                 end if
              end do
           end do
        end do

        ! Number of blocks with full ion2ion interactions
        num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,itype)
        iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,itype)
        do iblock = 1, num_block_full
           call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
                istart, iend, jstart, jend)
           do i = istart, iend
              ix = i
              iy = i + num_ions
              iz = i + num_ions*2
              do j = jstart, jend
                 jx = j
                 jy = j + num_ions
                 jz = j + num_ions*2
                 call minimum_image_displacement_@PBC@(a, b, c, &
                      xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                      xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                      dx, dy, dz, drnorm2)
                 if (drnorm2 < ewald%rcutsq) then
                    gmueij = gradmumelt_elec_ij(drnorm2, dx, dy, dz, &
                         alpha, alphasq, alphapi)
                    gradmu_efield(jx,ix) = gradmu_efield(jx,ix) + gmueij(1)
                    gradmu_efield(jy,ix) = gradmu_efield(jy,ix) + gmueij(2)
                    gradmu_efield(jz,ix) = gradmu_efield(jz,ix) + gmueij(3)

                    gradmu_efield(jy,iy) = gradmu_efield(jy,iy) + gmueij(4)
                    gradmu_efield(jz,iy) = gradmu_efield(jz,iy) + gmueij(5)

                    gradmu_efield(jz,iz) = gradmu_efield(jz,iz) + gmueij(6)
                 end if
              end do
           end do
        end do
     end if
  end do

end subroutine gradmumelt_electricfield_@PBC@

!================================================================================
! Compute the Coulomb potential felt by each electrode atoms due to all melt ions
!DEC$ ATTRIBUTES NOINLINE :: melt2elec_potential_@PBC@
subroutine mumelt2Qelec_potential_@PBC@(localwork, ewald, box, &
      ions, xyz_ions, electrodes, xyz_atoms, dipoles, V)
   implicit none

   ! Parameters in
   ! -------------
   type(MW_localwork_t), intent(in)    :: localwork !< Local work assignment
   type(MW_ewald_t),     intent(in)    :: ewald     !< Ewald summation parameters
   type(MW_box_t),       intent(in)    :: box       !< Simulation box parameters
   type(MW_ion_t),       intent(in)    :: ions(:)        !< ions parameters
   real(wp),             intent(in)    :: xyz_ions(:,:)  !< ions xyz positions
   type(MW_electrode_t), intent(in)    :: electrodes(:)   !< electrode parameters
   real(wp),             intent(in)    :: xyz_atoms(:,:)  !< electrode atoms xyz positions
   real(wp),             intent(in)    :: dipoles(:,:)

   ! Parameters out
   ! --------------
   real(wp),             intent(inout) :: V(:)       !< Coulomb potential

   ! Local
   ! -----
   integer  :: num_ion_types, num_elec_types
   integer  :: i, j, itype, jtype
   real(wp) :: a, b, c
   real(wp) :: qj !< ion charge
   real(wp) :: dx, dy, dz
   real(wp) :: drnorm2
   real(wp) :: alpha, alphasq, alphapi
   real(wp) :: etai, etaisq, etaipi
   real(wp) :: vi

   integer  :: num_block_full
   integer  :: iblock, iblock_offset, istart, iend, jstart, jend
   integer  :: count_n, offset_n, count_m, offset_m

   num_ion_types = size(ions,1)
   num_elec_types = size(electrodes,1)

   ! Precompute some factors
   ! -----------------------
   a = box%length(1)
   b = box%length(2)
   c = box%length(3)

   alpha = ewald%alpha
   alphasq = alpha*alpha
   alphapi = 2.0_wp*alpha/sqrtpi


   ! Short range contribution
   ! ------------------------
   do itype = 1, num_elec_types
      count_n = electrodes(itype)%count
      offset_n = electrodes(itype)%offset

      etai = electrodes(itype)%eta
      etaisq = etai*etai
      etaipi = 2.0_wp*etai/sqrtpi

      do jtype = 1, num_ion_types
         qj = ions(jtype)%charge
         count_m = ions(jtype)%count
         offset_m = ions(jtype)%offset

         if ((qj /= 0.0_wp).and.ewald%sr_interaction_ion2atom(itype,jtype)) then
            ! Number of blocks with full interactions (below the diagonal
            num_block_full = localwork%pair_ion2atom_num_block_full_local(itype,jtype)
            iblock_offset = localwork%pair_ion2atom_full_iblock_offset(itype,jtype)
            do iblock = 1, num_block_full
               call update_other_block_boundaries(iblock_offset+iblock, &
                     offset_n, count_n, offset_m, count_m, &
                     istart, iend, jstart, jend)
               do i = istart, iend
                  vi = 0.0_wp
                  do j = jstart, jend
                     call minimum_image_displacement_@PBC@(a, b, c, &
                           xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                           xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                           dx, dy, dz, drnorm2)
                     if (drnorm2 < ewald%rcutsq) then
                        vi = vi + pot_mumelt2Qelec_ij(drnorm2, dx, dy, dz, alpha, alphasq, alphapi, &
                              etai, etaisq, etaipi, dipoles(j,1), dipoles(j,2), dipoles(j,3))
                     end if
                  end do
                  V(i) = V(i) + vi
               end do
            end do
         end if
      end do
   end do
end subroutine mumelt2Qelec_potential_@PBC@

!================================================================================
! Compute the electric fields felt by each melt ions due to other melt ions
subroutine qmelt2mumelt_electricfield_@PBC@(localwork, ewald, box, ions, xyz_ions, &
     efield, tt)
  implicit none

  ! Parameters in
  ! -------------
  type(MW_localwork_t), intent(in)    :: localwork       !< Local work assignment
  type(MW_ewald_t),     intent(in)    :: ewald          !< Ewald summation parameters
  type(MW_box_t),       intent(in)    :: box            !< Simulation box parameters
  type(MW_ion_t),       intent(in)    :: ions(:)        !< ions parameters
  real(wp),             intent(in)    :: xyz_ions(:,:)  !< ions xyz positions
  type(MW_damping_t),   intent(in)    :: tt             !< Damping parameters

  ! Parameters out
  ! --------------
  real(wp),             intent(inout) :: efield(:,:) !< Electric field on ions

  ! Local
  ! -----
  integer :: num_ion_types
  integer :: i, j, itype, jtype
  integer :: kk
  real(wp) :: a, b, c
  real(wp) :: qi, qj !< ion charge
  real(wp) :: eij
  real(wp) :: drnorm, drnorm2, dx, dy, dz, drnorm3rec
  real(wp) :: alpha, alphasq, alphapi
  real(wp) :: xf, factorial
  real(wp) :: dampsumfi, dampexpi, dampfunci
  real(wp) :: dampsumfj, dampexpj, dampfuncj

  integer :: num_block_diag, num_block_full
  integer :: iblock, iblock_offset, istart, iend, jstart, jend
  integer :: count_n, offset_n, count_m, offset_m

  num_ion_types = size(ions,1)

  ! Precompute some factors
  ! -----------------------
  a = box%length(1)
  b = box%length(2)
  c = box%length(3)

  alpha = ewald%alpha
  alphasq = alpha*alpha
  alphapi = 2.0_wp*alpha/sqrtpi


  ! Short range contribution
  ! ------------------------
  do itype = 1, num_ion_types
     qi = ions(itype)%charge
     count_n = ions(itype)%count
     offset_n = ions(itype)%offset
     do jtype = 1, itype - 1
        count_m = ions(jtype)%count
        offset_m = ions(jtype)%offset
        qj = ions(jtype)%charge !PUT AFTER if? ADD (qj /= 0) in if condition?
        if (ewald%sr_interaction_ion2ion(itype,jtype)) then
           ! Number of blocks with full ion2ion interactions
           num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,jtype)
           iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,jtype)
           do iblock = 1, num_block_full
              call update_other_block_boundaries(iblock_offset+iblock, &
                   offset_n, count_n, offset_m, count_m, &
                   istart, iend, jstart, jend)
              do i = istart, iend
                 do j = jstart, jend
                    call minimum_image_displacement_@PBC@(a, b, c, &
                         xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                         xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                         dx, dy, dz, drnorm2)
                    if (drnorm2 < ewald%rcutsq) then
                       eij = elec_qmelt2mumelt_ij(drnorm2, alpha, alphasq, alphapi)

                       drnorm = dsqrt(drnorm2)
                       drnorm3rec = 1.0_wp/(drnorm*drnorm2)

                       dampsumfi = 1.0_wp
                       xf = 1.0_wp
                       factorial = 1.0_wp
                       do kk = 1, tt%order(jtype,itype)
                          xf = xf*tt%b(jtype,itype)*drnorm
                          factorial = factorial*float(kk)
                          dampsumfi = dampsumfi + (xf/factorial)
                       enddo
                       dampexpi = dexp(-tt%b(jtype,itype)*drnorm)
                       dampfunci = -dampsumfi*dampexpi*tt%c(jtype,itype)*drnorm3rec

                       efield(i,1) = efield(i,1) - qj*dx*(eij + dampfunci)
                       efield(i,2) = efield(i,2) - qj*dy*(eij + dampfunci)
                       efield(i,3) = efield(i,3) - qj*dz*(eij + dampfunci)

                       dampsumfj = 1.0_wp
                       xf = 1.0_wp
                       factorial = 1.0_wp
                       do kk = 1, tt%order(itype,jtype)
                          xf = xf*tt%b(itype,jtype)*drnorm
                          factorial = factorial*float(kk)
                          dampsumfj = dampsumfj + (xf/factorial)
                       enddo
                       dampexpj = dexp(-tt%b(itype,jtype)*drnorm)
                       dampfuncj = -dampsumfj*dampexpj*tt%c(itype,jtype)*drnorm3rec

                       efield(j,1) = efield(j,1) + qi*dx*(eij + dampfuncj)
                       efield(j,2) = efield(j,2) + qi*dy*(eij + dampfuncj)
                       efield(j,3) = efield(j,3) + qi*dz*(eij + dampfuncj)
                    end if
                 end do
              end do
           end do
        end if
     end do

     if (ewald%sr_interaction_ion2ion(itype,itype)) then
        ! Blocks on the diagonal compute only half of the pair interactions
        qj = ions(itype)%charge
        num_block_diag = localwork%pair_ion2ion_num_block_diag_local(itype)
        iblock_offset = localwork%pair_ion2ion_diag_iblock_offset(itype)

        do iblock = 1, num_block_diag
           call update_diag_block_boundaries(iblock_offset+iblock, &
                offset_n, count_n, istart, iend)
           do i = istart, iend
              do j = istart, iend
                 call minimum_image_displacement_@PBC@(a, b, c, &
                      xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                      xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                      dx, dy, dz, drnorm2)
                 if ((drnorm2 > 0.0_wp) .and. (drnorm2 < ewald%rcutsq)) then
                    eij = elec_qmelt2mumelt_ij(drnorm2, alpha, alphasq, alphapi)

                    drnorm = dsqrt(drnorm2)
                    drnorm3rec = 1.0_wp/(drnorm*drnorm2)
                    dampsumfj = 1.0_wp
                    xf = 1.0_wp
                    factorial = 1.0_wp
                    do kk = 1, tt%order(itype,itype)
                       xf = xf*tt%b(itype,itype)*drnorm
                       factorial = factorial*float(kk)
                       dampsumfj = dampsumfj + (xf/factorial)
                    enddo
                    dampexpj = dexp(-tt%b(itype,itype)*drnorm)
                    dampfuncj = -dampsumfj*dampexpj*tt%c(itype,itype)*drnorm3rec

                    efield(j,1) = efield(j,1) + qi*dx*(eij + dampfuncj)
                    efield(j,2) = efield(j,2) + qi*dy*(eij + dampfuncj)
                    efield(j,3) = efield(j,3) + qi*dz*(eij + dampfuncj)
                 end if
              end do
           end do
        end do

        ! Number of blocks with full ion2ion interactions
        qi = ions(itype)%charge !NECESSARY?
        qj = ions(itype)%charge
        num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,itype)
        iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,itype)
        do iblock = 1, num_block_full
           call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
                istart, iend, jstart, jend)
           do i = istart, iend
              do j = jstart, jend
                 call minimum_image_displacement_@PBC@(a, b, c, &
                      xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                      xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                      dx, dy, dz, drnorm2)
                 if (drnorm2 < ewald%rcutsq) then
                    eij = elec_qmelt2mumelt_ij(drnorm2, alpha, alphasq, alphapi)

                    drnorm = dsqrt(drnorm2)
                    drnorm3rec = 1.0_wp/(drnorm*drnorm2)
                    dampsumfj = 1.0_wp
                    xf = 1.0_wp
                    factorial = 1.0_wp
                    do kk = 1, tt%order(itype,itype)
                       xf = xf*tt%b(itype,itype)*drnorm
                       factorial = factorial*float(kk)
                       dampsumfj = dampsumfj + (xf/factorial)
                    enddo
                    dampexpj = dexp(-tt%b(itype,itype)*drnorm)
                    dampfuncj = -dampsumfj*dampexpj*tt%c(itype,itype)*drnorm3rec

                    efield(i,1) = efield(i,1) - qj*dx*(eij + dampfuncj)
                    efield(i,2) = efield(i,2) - qj*dy*(eij + dampfuncj)
                    efield(i,3) = efield(i,3) - qj*dz*(eij + dampfuncj)

                    efield(j,1) = efield(j,1) + qi*dx*(eij + dampfuncj)
                    efield(j,2) = efield(j,2) + qi*dy*(eij + dampfuncj)
                    efield(j,3) = efield(j,3) + qi*dz*(eij + dampfuncj)

                 end if
              end do
           end do
        end do
     end if
  end do

end subroutine qmelt2mumelt_electricfield_@PBC@

!================================================================================
! Compute the electric fields felt by each melt ions due to other melt ions
subroutine mumelt2mumelt_electricfield_@PBC@(localwork, ewald, box, ions, xyz_ions, &
     efield, dipoles)
  implicit none

  ! Parameters in
  ! -------------
  type(MW_localwork_t), intent(in)    :: localwork       !< Local work assignment
  type(MW_ewald_t),     intent(in)    :: ewald          !< Ewald summation parameters
  type(MW_box_t),       intent(in)    :: box            !< Simulation box parameters
  type(MW_ion_t),       intent(in)    :: ions(:)        !< ions parameters
  real(wp),             intent(in)    :: xyz_ions(:,:)  !< ions xyz positions
  real(wp),             intent(in)    :: dipoles(:,:)  !< ions dipoles

  ! Parameters out
  ! --------------
  real(wp),             intent(inout) :: efield(:,:) !< Electric field on ions

  ! Local
  ! -----
  integer :: num_ion_types
  integer :: i, j, itype, jtype
  real(wp) :: a, b, c
  real(wp) :: muix, muiy, muiz, mujx, mujy, mujz !< ion dipole
  real(wp) :: drnorm2, dx, dy, dz
  real(wp) :: alpha, alphasq, alphapi
  real(wp), dimension(6) :: eij

  integer :: num_block_diag, num_block_full
  integer :: iblock, iblock_offset, istart, iend, jstart, jend
  integer :: count_n, offset_n, count_m, offset_m

  num_ion_types = size(ions,1)

  ! Precompute some factors
  ! -----------------------
  a = box%length(1)
  b = box%length(2)
  c = box%length(3)

  alpha = ewald%alpha
  alphasq = alpha*alpha
  alphapi = 2.0_wp*alpha/sqrtpi


  ! Short range contribution
  ! ------------------------
  do itype = 1, num_ion_types
     count_n = ions(itype)%count
     offset_n = ions(itype)%offset
     do jtype = 1, itype - 1
        count_m = ions(jtype)%count
        offset_m = ions(jtype)%offset
        if (ewald%sr_interaction_ion2ion(itype,jtype)) then
           ! Number of blocks with full ion2ion interactions
           num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,jtype)
           iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,jtype)
           do iblock = 1, num_block_full
              call update_other_block_boundaries(iblock_offset+iblock, &
                   offset_n, count_n, offset_m, count_m, &
                   istart, iend, jstart, jend)
              do i = istart, iend
                 muix = dipoles(i,1)
                 muiy = dipoles(i,2)
                 muiz = dipoles(i,3)
                 do j = jstart, jend
                    call minimum_image_displacement_@PBC@(a, b, c, &
                         xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                         xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                         dx, dy, dz, drnorm2)
                    if (drnorm2 < ewald%rcutsq) then
                       mujx = dipoles(j,1)
                       mujy = dipoles(j,2)
                       mujz = dipoles(j,3)
                       eij = elec_mumelt2mumelt_ij(drnorm2, dx, dy, dz, &
                            alpha, alphasq, alphapi, &
                            muix, muiy, muiz, mujx, mujy, mujz)
                       efield(i,1) = efield(i,1) + eij(4)
                       efield(i,2) = efield(i,2) + eij(5)
                       efield(i,3) = efield(i,3) + eij(6)

                       efield(j,1) = efield(j,1) + eij(1)
                       efield(j,2) = efield(j,2) + eij(2)
                       efield(j,3) = efield(j,3) + eij(3)
                    end if
                 end do
              end do
           end do
        end if
     end do

     if (ewald%sr_interaction_ion2ion(itype,itype)) then
        ! Blocks on the diagonal compute only half of the pair interactions
        num_block_diag = localwork%pair_ion2ion_num_block_diag_local(itype)
        iblock_offset = localwork%pair_ion2ion_diag_iblock_offset(itype)

        do iblock = 1, num_block_diag
           call update_diag_block_boundaries(iblock_offset+iblock, &
                offset_n, count_n, istart, iend)
           do i = istart, iend
              muix = dipoles(i,1)
              muiy = dipoles(i,2)
              muiz = dipoles(i,3)
              do j = istart, iend
                 call minimum_image_displacement_@PBC@(a, b, c, &
                      xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                      xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                      dx, dy, dz, drnorm2)
                 if ((drnorm2 > 0.0_wp) .and. (drnorm2 < ewald%rcutsq)) then
                    mujx = dipoles(j,1)
                    mujy = dipoles(j,2)
                    mujz = dipoles(j,3)
                    eij = elec_mumelt2mumelt_ij(drnorm2, dx, dy, dz, &
                         alpha, alphasq, alphapi, &
                         muix, muiy, muiz, mujx, mujy, mujz)
                    efield(j,1) = efield(j,1) + eij(1)
                    efield(j,2) = efield(j,2) + eij(2)
                    efield(j,3) = efield(j,3) + eij(3)
                 end if
              end do
           end do
        end do

        ! Number of blocks with full ion2ion interactions
        num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,itype)
        iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,itype)
        do iblock = 1, num_block_full
           call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
                istart, iend, jstart, jend)
           do i = istart, iend
              muix = dipoles(i,1)
              muiy = dipoles(i,2)
              muiz = dipoles(i,3)
              do j = jstart, jend
                 call minimum_image_displacement_@PBC@(a, b, c, &
                      xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                      xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                      dx, dy, dz, drnorm2)
                 if (drnorm2 < ewald%rcutsq) then
                    mujx = dipoles(j,1)
                    mujy = dipoles(j,2)
                    mujz = dipoles(j,3)
                    eij = elec_mumelt2mumelt_ij(drnorm2, dx, dy, dz, &
                         alpha, alphasq, alphapi, &
                         muix, muiy, muiz, mujx, mujy, mujz)

                    efield(i,1) = efield(i,1) + eij(4)
                    efield(i,2) = efield(i,2) + eij(5)
                    efield(i,3) = efield(i,3) + eij(6)

                    efield(j,1) = efield(j,1) + eij(1)
                    efield(j,2) = efield(j,2) + eij(2)
                    efield(j,3) = efield(j,3) + eij(3)
                 end if
              end do
           end do
        end do
     end if
  end do
end subroutine mumelt2mumelt_electricfield_@PBC@

!================================================================================
! Compute the electric fields felt by each melt ions due to other melt ions
subroutine Qelec2mumelt_electricfield_@PBC@(localwork, ewald, box, ions, xyz_ions, &
     electrodes, xyz_atoms, q_elec, efield)
  implicit none

  ! Parameters in
  ! -------------
  type(MW_localwork_t), intent(in)    :: localwork       !< Local work assignment
  type(MW_ewald_t),     intent(in)    :: ewald          !< Ewald summation parameters
  type(MW_box_t),       intent(in)    :: box            !< Simulation box parameters
  type(MW_ion_t),       intent(in)    :: ions(:)        !< ions parameters
  real(wp),             intent(in)    :: xyz_ions(:,:)  !< ions xyz positions
  type(MW_electrode_t), intent(in)    :: electrodes(:)   !< electrode parameters
  real(wp),             intent(in)    :: xyz_atoms(:,:)  !< electrode atoms xyz positions
  real(wp),             intent(in)    :: q_elec(:)      !< Electrode atoms charge

  ! Parameters out
  ! --------------
  real(wp),             intent(inout) :: efield(:,:) !< Electric field on ions

  ! Local
  ! -----
  integer :: num_ion_types, num_elec_types
  integer :: i, j, itype, jtype
  real(wp) :: a, b, c
  real(wp) :: qi !< ion charge
  real(wp) :: Q_j !< electrode charge
  real(wp) :: drnorm2, dx, dy, dz
  real(wp) :: alpha, alphasq, alphapi
  real(wp) :: etaj, etajsq, etajpi
  real(wp), dimension(3) :: eij

  integer :: num_block_full
  integer :: iblock, iblock_offset, istart, iend, jstart, jend
  integer :: count_n, offset_n, count_m, offset_m

  num_ion_types = size(ions,1)
  num_elec_types = size(electrodes,1)

  ! Precompute some factors
  ! -----------------------
  a = box%length(1)
  b = box%length(2)
  c = box%length(3)

  alpha = ewald%alpha
  alphasq = alpha*alpha
  alphapi = 2.0_wp*alpha/sqrtpi


  ! Short range contribution
  ! ------------------------
  do itype = 1, num_ion_types
     qi = ions(itype)%charge
     count_n = ions(itype)%count
     offset_n = ions(itype)%offset
     do jtype = 1, num_elec_types
        if (ewald%sr_interaction_atom2ion(itype,jtype)) then
           count_m = electrodes(jtype)%count
           offset_m = electrodes(jtype)%offset
           ! Number of blocks with full ion2ion interactions
           num_block_full = localwork%pair_atom2ion_num_block_full_local(itype,jtype)
           iblock_offset = localwork%pair_atom2ion_full_iblock_offset(itype,jtype)
           etaj = electrodes(jtype)%eta
           etajsq = etaj*etaj
           etajpi = 2.0_wp*etaj/sqrtpi
           do iblock = 1, num_block_full
              call update_other_block_boundaries(iblock_offset+iblock, &
                   offset_n, count_n, offset_m, count_m, &
                   istart, iend, jstart, jend)
              do i = istart, iend
                 do j = jstart, jend
                    call minimum_image_displacement_@PBC@(a, b, c, &
                         xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                         xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                         dx, dy, dz, drnorm2)
                    if (drnorm2 < ewald%rcutsq) then
                       Q_j = q_elec(j)
                       eij = elec_Qelec2mumelt_ij(drnorm2, dx, dy, dz, &
                           alpha, alphasq, alphapi, etaj, etajsq, etajpi, Q_j)
                       efield(i,1) = efield(i,1) + eij(1)
                       efield(i,2) = efield(i,2) + eij(2)
                       efield(i,3) = efield(i,3) + eij(3)
                    end if
                 end do
              end do
              !acc end loop
           end do
        end if
     end do
  end do
end subroutine Qelec2mumelt_electricfield_@PBC@

!================================================================================
! Compute the Coulomb forces felt by each melt ions due to other melt ions
subroutine melt_forces_pim_@PBC@(localwork, ewald, box, ions, xyz_ions, &
     electrodes, xyz_atoms, q_elec, force, stress_tensor, dipoles, tt)
  implicit none

  ! Parameters in
  ! -------------
  type(MW_localwork_t), intent(in)    :: localwork       !< Local work assignment
  type(MW_ewald_t),     intent(in)    :: ewald          !< Ewald summation parameters
  type(MW_box_t),       intent(in)    :: box            !< Simulation box parameters
  type(MW_ion_t),       intent(in)    :: ions(:)        !< ions parameters
  real(wp),             intent(in)    :: xyz_ions(:,:)  !< ions xyz positions
  type(MW_electrode_t), intent(in)    :: electrodes(:)   !< electrode parameters
  real(wp),             intent(in)    :: xyz_atoms(:,:)  !< electrode atoms xyz positions
  real(wp),             intent(in)    :: q_elec(:)      !< Electrode atoms charge
  real(wp),             intent(in)    :: dipoles(:,:)  !< ions dipoles
  type(MW_damping_t),   intent(in)    :: tt             !< Damping parameters

  ! Parameters out
  ! --------------
  real(wp),             intent(inout) :: force(:,:) !< Coulomb force on ions
  real(wp),             intent(inout) :: stress_tensor(:,:) !< Coulomb force contribution to stress tensor

  ! Local
  ! -----
  integer :: num_ion_types, num_elec_types
  integer :: i, j, itype, jtype
  integer :: orderij, orderji
  real(wp) :: a, b, c
  real(wp) :: qi, qj !< ion charge
  real(wp) :: muix, muiy, muiz, mujx, mujy, mujz !< ion dipole
  real(wp) :: Q_j !< electrode charges
  real(wp), dimension(3) :: fij
  real(wp) :: drnorm2, dx, dy, dz
  real(wp) :: alpha, alphasq, alphapi
  real(wp) :: etaj, etajsq, etajpi
  real(wp) :: bij, bji, cij, cji

  integer :: num_block_diag, num_block_full
  integer :: iblock, iblock_offset, istart, iend, jstart, jend
  integer :: count_n, offset_n, count_m, offset_m

  num_ion_types = size(ions,1)
  num_elec_types = size(electrodes,1)

  ! Precompute some factors
  ! -----------------------
  a = box%length(1)
  b = box%length(2)
  c = box%length(3)

  alpha = ewald%alpha
  alphasq = alpha*alpha
  alphapi = 2.0_wp*alpha/sqrtpi


  ! Short range contribution
  ! ------------------------
  do itype = 1, num_ion_types
     count_n = ions(itype)%count
     offset_n = ions(itype)%offset
     qi = ions(itype)%charge
     if (qi /= 0.0_wp) then
        do jtype = 1, itype - 1
           count_m = ions(jtype)%count
           offset_m = ions(jtype)%offset
           qj = ions(jtype)%charge
           if ((qj /= 0.0_wp) .and. ewald%sr_interaction_ion2ion(itype,jtype)) then
              ! Number of blocks with full ion2ion interactions
              num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,jtype)
              iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,jtype)
              do iblock = 1, num_block_full
                 call update_other_block_boundaries(iblock_offset+iblock, &
                      offset_n, count_n, offset_m, count_m, &
                      istart, iend, jstart, jend)
                 do i = istart, iend
                    muix = dipoles(i,1)
                    muiy = dipoles(i,2)
                    muiz = dipoles(i,3)
                    do j = jstart, jend
                       call minimum_image_displacement_@PBC@(a, b, c, &
                            xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                            xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                            dx, dy, dz, drnorm2)
                       if (drnorm2 < ewald%rcutsq) then
                          mujx = dipoles(j,1)
                          mujy = dipoles(j,2)
                          mujz = dipoles(j,3)

                          bij = tt%b(itype,jtype)
                          bji = tt%b(jtype,itype)
                          orderij = tt%order(itype,jtype)
                          orderji = tt%order(jtype,itype)
                          cij = tt%c(itype,jtype)
                          cji = tt%c(jtype,itype)

                          fij = force_melt2melt_ij(drnorm2, dx, dy, dz,&
                               alpha, alphasq, alphapi, qi, qj,&
                               muix, muiy, muiz, mujx, mujy, mujz, &
                               bij, bji, orderij, orderji, cij, cji)
                          force(i,1) = force(i,1) + fij(1)
                          force(i,2) = force(i,2) + fij(2)
                          force(i,3) = force(i,3) + fij(3)

                          force(j,1) = force(j,1) - fij(1)
                          force(j,2) = force(j,2) - fij(2)
                          force(j,3) = force(j,3) - fij(3)

			   stress_tensor(1,1)=stress_tensor(1,1)- dx * fij(1)
			   stress_tensor(2,2)=stress_tensor(2,2)- dy * fij(2)
			   stress_tensor(3,3)=stress_tensor(3,3)- dz * fij(3)
			   stress_tensor(1,2)=stress_tensor(1,2)- 0.5_wp * (dx * fij(2) + dy * fij(1))
			   stress_tensor(1,3)=stress_tensor(1,3)- 0.5_wp * (dx * fij(3) + dz * fij(1))
			   stress_tensor(2,3)=stress_tensor(2,3)- 0.5_wp * (dy * fij(3) + dz * fij(2))
                       end if
                    end do
                 end do
              end do
           end if
        end do

        if (ewald%sr_interaction_ion2ion(itype,itype)) then
           ! Blocks on the diagonal compute only half of the pair interactions
           qj = ions(itype)%charge
           num_block_diag = localwork%pair_ion2ion_num_block_diag_local(itype)
           iblock_offset = localwork%pair_ion2ion_diag_iblock_offset(itype)

           bij = tt%b(itype,itype)
           bji = tt%b(itype,itype)
           orderij = tt%order(itype,itype)
           orderji = tt%order(itype,itype)
           cij = tt%c(itype,itype)
           cji = tt%c(itype,itype)

           do iblock = 1, num_block_diag
              call update_diag_block_boundaries(iblock_offset+iblock, &
                   offset_n, count_n, istart, iend)
              do i = istart, iend
                 muix = dipoles(i,1)
                 muiy = dipoles(i,2)
                 muiz = dipoles(i,3)
                 do j = istart, iend
                    call minimum_image_displacement_@PBC@(a, b, c, &
                         xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                         xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                         dx, dy, dz, drnorm2)
                    if ((drnorm2 > 0.0_wp) .and. (drnorm2 < ewald%rcutsq)) then
                       mujx = dipoles(j,1)
                       mujy = dipoles(j,2)
                       mujz = dipoles(j,3)

                       fij = force_melt2melt_ij(drnorm2, dx, dy, dz,&
                            alpha, alphasq, alphapi, qi, qj,&
                            muix, muiy, muiz, mujx, mujy, mujz,&
                            bij, bji, orderij, orderji, cij, cji)
                       force(j,1) = force(j,1) - fij(1)
                       force(j,2) = force(j,2) - fij(2)
                       force(j,3) = force(j,3) - fij(3)

		       stress_tensor(1,1)=stress_tensor(1,1)- 0.5_wp * dx * fij(1)
		       stress_tensor(2,2)=stress_tensor(2,2)- 0.5_wp * dy * fij(2)
		       stress_tensor(3,3)=stress_tensor(3,3)- 0.5_wp * dz * fij(3)
		       stress_tensor(1,2)=stress_tensor(1,2)- 0.25_wp * (dx * fij(2) + dy * fij(1))
		       stress_tensor(1,3)=stress_tensor(1,3)- 0.25_wp * (dx * fij(3) + dz * fij(1))
		       stress_tensor(2,3)=stress_tensor(2,3)- 0.25_wp * (dy * fij(3) + dz * fij(2))
                    end if
                 end do
              end do
           end do

           ! Number of blocks with full ion2ion interactions
           qi = ions(itype)%charge
           qj = ions(itype)%charge
           num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,itype)
           iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,itype)
           do iblock = 1, num_block_full
              call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
                   istart, iend, jstart, jend)
              do i = istart, iend
                 muix = dipoles(i,1)
                 muiy = dipoles(i,2)
                 muiz = dipoles(i,3)
                 do j = jstart, jend
                    call minimum_image_displacement_@PBC@(a, b, c, &
                         xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                         xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                         dx, dy, dz, drnorm2)
                    if (drnorm2 < ewald%rcutsq) then
                       mujx = dipoles(j,1)
                       mujy = dipoles(j,2)
                       mujz = dipoles(j,3)
                       fij = force_melt2melt_ij(drnorm2, dx, dy, dz,&
                            alpha, alphasq, alphapi, qi, qj,&
                            muix, muiy, muiz, mujx, mujy, mujz,&
                            bij, bji, orderij, orderji, cij, cji)

                       force(i,1) = force(i,1) + fij(1)
                       force(i,2) = force(i,2) + fij(2)
                       force(i,3) = force(i,3) + fij(3)

                       force(j,1) = force(j,1) - fij(1)
                       force(j,2) = force(j,2) - fij(2)
                       force(j,3) = force(j,3) - fij(3)

		       stress_tensor(1,1)=stress_tensor(1,1)- dx * fij(1)
		       stress_tensor(2,2)=stress_tensor(2,2)- dy * fij(2)
		       stress_tensor(3,3)=stress_tensor(3,3)- dz * fij(3)
		       stress_tensor(1,2)=stress_tensor(1,2)- 0.5_wp * (dx * fij(2) + dy * fij(1))
		       stress_tensor(1,3)=stress_tensor(1,3)- 0.5_wp * (dx * fij(3) + dz * fij(1))
		       stress_tensor(2,3)=stress_tensor(2,3)- 0.5_wp * (dy * fij(3) + dz * fij(2))
                    end if
                 end do
              end do
           end do
        end if

        qi = ions(itype)%charge
        do jtype = 1, num_elec_types
           if (ewald%sr_interaction_atom2ion(itype,jtype)) then
              count_m = electrodes(jtype)%count
              offset_m = electrodes(jtype)%offset
              ! Number of blocks with full ion2ion interactions
              num_block_full = localwork%pair_atom2ion_num_block_full_local(itype,jtype)
              iblock_offset = localwork%pair_atom2ion_full_iblock_offset(itype,jtype)
              etaj = electrodes(jtype)%eta
              etajsq = etaj*etaj
              etajpi = 2.0_wp*etaj/sqrtpi
              do iblock = 1, num_block_full
                 call update_other_block_boundaries(iblock_offset+iblock, &
                      offset_n, count_n, offset_m, count_m, &
                      istart, iend, jstart, jend)
                 do i = istart, iend
                    muix = dipoles(i,1)
                    muiy = dipoles(i,2)
                    muiz = dipoles(i,3)
                    do j = jstart, jend
                       call minimum_image_displacement_@PBC@(a, b, c, &
                            xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                            xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                            dx, dy, dz, drnorm2)
                       if (drnorm2 < ewald%rcutsq) then
                          Q_j = q_elec(j)
                          fij = force_elec2melt_ij(drnorm2, dx, dy, dz, &
                              alpha, alphasq, alphapi, etaj, etajsq, etajpi, &
                              qi, Q_j, muix, muiy, muiz)

                          force(i,1) = force(i,1) + fij(1)
                          force(i,2) = force(i,2) + fij(2)
                          force(i,3) = force(i,3) + fij(3)

		          stress_tensor(1,1)=stress_tensor(1,1)- dx * fij(1)
		          stress_tensor(2,2)=stress_tensor(2,2)- dy * fij(2)
		          stress_tensor(3,3)=stress_tensor(3,3)- dz * fij(3)
		          stress_tensor(1,2)=stress_tensor(1,2)- 0.5_wp * (dx * fij(2) + dy * fij(1))
		          stress_tensor(1,3)=stress_tensor(1,3)- 0.5_wp * (dx * fij(3) + dz * fij(1))
		          stress_tensor(2,3)=stress_tensor(2,3)- 0.5_wp * (dy * fij(3) + dz * fij(2))
                       end if
                    end do
                 end do
                 !acc end loop
              end do
           end if
        end do
     end if
  end do
end subroutine melt_forces_pim_@PBC@

!================================================================================
! Compute the Energy due to Coulomb interactions
subroutine energy_pim_@PBC@(localwork, ewald, box, ions, xyz_ions, &
     electrodes, xyz_atoms, q_elec, h, dipoles, tt)
  use MW_constants, only: sqrt2
  implicit none

  ! Parameters in
  ! -------------
  type(MW_localwork_t), intent(in)    :: localwork       !< Local work assignment
  type(MW_ewald_t),     intent(in)    :: ewald          !< Ewald summation parameters
  type(MW_box_t),       intent(in)    :: box            !< Simulation box parameters
  type(MW_ion_t),       intent(in)    :: ions(:)        !< ions parameters
  real(wp),             intent(in)    :: xyz_ions(:,:)  !< ions xyz positions
  type(MW_electrode_t), intent(in)    :: electrodes(:)   !< electrode parameters
  real(wp),             intent(in)    :: xyz_atoms(:,:)  !< electrode atoms xyz positions
  real(wp),             intent(in)    :: q_elec(:)      !< Electrode atoms charge
  real(wp),             intent(in)    :: dipoles(:,:)  !< ions dipoles
  type(MW_damping_t),   intent(in)    :: tt             !< Damping parameters

  ! Parameters out
  ! --------------
  real(wp),             intent(inout) :: h !< Coulomb energy

  ! Local
  ! -----
  integer :: num_ion_types, num_elec_types
  integer :: i, j, itype, jtype
  integer :: orderij, orderji
  real(wp) :: a, b, c
  real(wp) :: qi, qj !< ion charge
  real(wp) :: Q_i, Q_j !< electrode charges
  real(wp) :: muix, muiy, muiz
  real(wp) :: mujx, mujy, mujz
  real(wp) :: bij, bji, cij, cji
  real(wp) :: erfdiff
  real(wp) :: vi, pot_ij
  real(wp) :: drnorm, drnorm2, rcutsq
  real(wp) :: dx, dy, dz
  real(wp) :: alpha, alphasq, alphapi
  real(wp) :: enerij
  real(wp) :: etai, etaj, etajsq, etajpi, etaij, etaijsq

  integer :: num_block_diag, num_block_full
  integer :: iblock, iblock_offset, istart, iend, jstart, jend
  integer :: count_n, offset_n, count_m, offset_m

  num_ion_types = size(ions,1)
  num_elec_types = size(electrodes,1)

  ! Precompute some factors
  ! -----------------------
  a = box%length(1)
  b = box%length(2)
  c = box%length(3)

  alpha = ewald%alpha
  alphasq = alpha*alpha
  alphapi = 2.0_wp*alpha/sqrtpi

  rcutsq = ewald%rcutsq
  ! Zero energy
  ! ----------------
  h = 0.0_wp
  vi = 0.0_wp


  ! Short range contribution from Melt->Melt and Elec->Melt interaction
  ! -------------------------------------------------------------------
  do itype = 1, num_ion_types
     count_n = ions(itype)%count
     offset_n = ions(itype)%offset
     qi = ions(itype)%charge
     if (qi /= 0.0_wp) then
        !Here we compute qq, qmu and mumu interactions
        do jtype = 1, itype - 1
           count_m = ions(jtype)%count
           offset_m = ions(jtype)%offset
           qj = ions(jtype)%charge
           if ((qj /= 0.0_wp) .and. ewald%sr_interaction_ion2ion(itype,jtype)) then
              ! Number of blocks with full ion2ion interactions
              num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,jtype)
              iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,jtype)

              do iblock = 1, num_block_full
                 call update_other_block_boundaries(iblock+iblock_offset, &
                      offset_n, count_n, offset_m, count_m, &
                      istart, iend, jstart, jend)
                 do i = istart, iend
                    muix = dipoles(i,1)
                    muiy = dipoles(i,2)
                    muiz = dipoles(i,3)
                    do j = jstart, jend
                       call minimum_image_displacement_@PBC@(a, b, c, &
                            xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                            xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                            dx, dy, dz, drnorm2)
                       if (drnorm2 < rcutsq) then
                          mujx = dipoles(j,1)
                          mujy = dipoles(j,2)
                          mujz = dipoles(j,3)
                          bij = tt%b(itype,jtype)
                          bji = tt%b(jtype,itype)
                          orderij = tt%order(itype,jtype)
                          orderji = tt%order(jtype,itype)
                          cij = tt%c(itype,jtype)
                          cji = tt%c(jtype,itype)
                          enerij = energy_melt2melt_ij(drnorm2, dx, dy, dz, &
                               alpha, alphasq, alphapi, qi, qj, &
                               muix, muiy, muiz, mujx, mujy, mujz, &
                               bij, bji, orderij, orderji, cij, cji)
                          vi = vi + enerij
                       end if
                    end do
                 end do
              end do
           end if
        end do

        if (ewald%sr_interaction_ion2ion(itype,itype)) then
           ! Blocks on the diagonal compute only half of the pair interactions
           qj = ions(itype)%charge
           num_block_diag = localwork%pair_ion2ion_num_block_diag_local(itype)
           iblock_offset = localwork%pair_ion2ion_diag_iblock_offset(itype)

           bij = tt%b(itype,itype)
           bji = tt%b(itype,itype)
           orderij = tt%order(itype,itype)
           orderji = tt%order(itype,itype)
           cij = tt%c(itype,itype)
           cji = tt%c(itype,itype)

           do iblock = 1, num_block_diag
              call update_diag_block_boundaries(iblock_offset+iblock, offset_n, count_n, istart, iend)
              do i = istart, iend
                 muix = dipoles(i,1)
                 muiy = dipoles(i,2)
                 muiz = dipoles(i,3)
                 do j = istart, iend
                    call minimum_image_displacement_@PBC@(a, b, c, &
                         xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                         xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                         dx, dy, dz, drnorm2)
                    if ((drnorm2 > 0.0_wp) .and. (drnorm2 < rcutsq)) then
                       mujx = dipoles(j,1)
                       mujy = dipoles(j,2)
                       mujz = dipoles(j,3)
                       enerij = energy_melt2melt_ij(drnorm2, dx, dy, dz, &
                            alpha, alphasq, alphapi, qi, qj, &
                            muix, muiy, muiz, mujx, mujy, mujz, &
                            bij, bji, orderij, orderji, cij, cji)
                       vi = vi + 0.5_wp*enerij
                    end if
                 end do
              end do
           end do

           qi = ions(itype)%charge
           qj = ions(itype)%charge
           ! Number of blocks with full ion2ion interactions
           num_block_full = localwork%pair_ion2ion_num_block_full_local(itype,itype)
           iblock_offset = localwork%pair_ion2ion_full_iblock_offset(itype,itype)

           do iblock = 1, num_block_full
              call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
                   istart, iend, jstart, jend)
              do i = istart, iend
                 muix = dipoles(i,1)
                 muiy = dipoles(i,2)
                 muiz = dipoles(i,3)
                 do j = jstart, jend
                    call minimum_image_displacement_@PBC@(a, b, c, &
                         xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                         xyz_ions(j,1), xyz_ions(j,2), xyz_ions(j,3), &
                         dx, dy, dz, drnorm2)
                    if (drnorm2 < rcutsq) then
                       mujx = dipoles(j,1)
                       mujy = dipoles(j,2)
                       mujz = dipoles(j,3)
                       enerij = energy_melt2melt_ij(drnorm2, dx, dy, dz, &
                            alpha, alphasq, alphapi, qi, qj, &
                            muix, muiy, muiz, mujx, mujy, mujz, &
                            bij, bji, orderij, orderji, cij, cji)
                       vi = vi + enerij
                    end if
                 end do
              end do
           end do
        end if

        !Here we compute qQ and Qmu interaction
        !We need damping functions for Qmu interaction.
        qi = ions(itype)%charge
        do jtype = 1, num_elec_types
           count_m = electrodes(jtype)%count
           offset_m = electrodes(jtype)%offset
           etaj = electrodes(jtype)%eta
           etajsq = etaj*etaj
           etajpi = 2.0_wp*etaj/sqrtpi
           if (ewald%sr_interaction_atom2ion(itype,jtype)) then
              ! Number of blocks with full ion2ion interactions
              num_block_full = localwork%pair_atom2ion_num_block_full_local(itype,jtype)
              iblock_offset = localwork%pair_atom2ion_full_iblock_offset(itype,jtype)

              do iblock = 1, num_block_full
                 call update_other_block_boundaries(iblock+iblock_offset, &
                      offset_n, count_n, offset_m, count_m, &
                      istart, iend, jstart, jend)
                 do i = istart, iend
                    muix = dipoles(i,1)
                    muiy = dipoles(i,2)
                    muiz = dipoles(i,3)
                    do j = jstart, jend
                       call minimum_image_displacement_@PBC@(a, b, c, &
                            xyz_ions(i,1), xyz_ions(i,2), xyz_ions(i,3), &
                            xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                            dx, dy, dz, drnorm2)
                       if (drnorm2 < rcutsq) then
                          Q_j = q_elec(j)
                          vi = vi + energy_elec2melt_ij(drnorm2, dx, dy, dz, &
                              alpha, alphasq, alphapi, etaj, etajsq, etajpi, &
                              qi, Q_j, muix, muiy, muiz, bji, orderji, 0.0_wp)
                       end if
                    end do
                 end do
              end do
           end if
        end do
     end if
  end do

  ! Elec -> Elec interactions
  !Here we compute QQ interactions
  do itype = 1, num_elec_types
     count_n = electrodes(itype)%count
     offset_n = electrodes(itype)%offset
     do jtype = 1, itype-1
        count_m =  electrodes(jtype)%count
        offset_m = electrodes(jtype)%offset
        if (ewald%sr_interaction_atom2atom(itype,jtype)) then
           etai = electrodes(itype)%eta
           etaj = electrodes(jtype)%eta
           etaij = (etai*etaj)/sqrt(etai*etai + etaj*etaj)
           etaijsq = etaij * etaij

           ! Number of blocks with full interactions (below the diagonal
           num_block_full = localwork%pair_atom2atom_num_block_full_local(itype,jtype)
           iblock_offset = localwork%pair_atom2atom_full_iblock_offset(itype,jtype)

           do iblock = 1, num_block_full
              call update_other_block_boundaries(iblock_offset+iblock, &
                   offset_n, count_n, offset_m, count_m, istart, iend, jstart, jend)
              do i = istart, iend
                 Q_i = q_elec(i)
                 do j = jstart, jend
                    call minimum_image_distance_@PBC@(a, b, c, &
                         xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                         xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                         drnorm2)
                    if (drnorm2 < rcutsq)  then
                       Q_j = q_elec(j)
                       drnorm = sqrt(drnorm2)
                       ! Change erfc to 1-erf because of underflows
                       erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                       pot_ij = erfdiff/drnorm
                       Vi = Vi + Q_i*Q_j*pot_ij
                    end if
                 end do
              end do
           end do
        end if
     end do

     if (ewald%sr_interaction_atom2atom(itype,itype)) then

        etai = electrodes(itype)%eta
        etaij = etai/sqrt2
        etaijsq = etaij*etaij

        ! Blocks on the diagonal compute only half of the pair interactions
        num_block_diag = localwork%pair_atom2atom_num_block_diag_local(itype)
        iblock_offset = localwork%pair_atom2atom_diag_iblock_offset(itype)

        do iblock = 1, num_block_diag
           call update_diag_block_boundaries(iblock+iblock_offset, &
                offset_n, count_n, istart, iend)
           do i = istart, iend
             Q_i = q_elec(i)
              do j = istart, iend
                 call minimum_image_distance_@PBC@(a, b, c, &
                      xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                      xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                      drnorm2)
                 if ((drnorm2 > 0.0_wp) .and. (drnorm2 < rcutsq))  then
                    Q_j = q_elec(j)
                    drnorm = sqrt(drnorm2)
                    erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                    pot_ij =  erfdiff/drnorm
                    vi = vi + 0.5_wp*Q_i*Q_j*pot_ij
                 end if
              end do
           end do
        end do

        ! Number of blocks with full interactions (below the diagonal
        num_block_full = localwork%pair_atom2atom_num_block_full_local(itype,itype)
        iblock_offset = localwork%pair_atom2atom_full_iblock_offset(itype,itype)

        do iblock = 1, num_block_full
           call update_tri_block_boundaries(iblock_offset+iblock, offset_n, count_n, &
                istart, iend, jstart, jend)
           do i = istart, iend
             Q_i = q_elec(i)
              do j = jstart, jend
                 call minimum_image_distance_@PBC@(a, b, c, &
                      xyz_atoms(i,1), xyz_atoms(i,2), xyz_atoms(i,3), &
                      xyz_atoms(j,1), xyz_atoms(j,2), xyz_atoms(j,3), &
                      drnorm2)
                 if (drnorm2 < rcutsq)  then
                    Q_j = q_elec(j)
                    drnorm = sqrt(drnorm2)
                    erfdiff = (erfc(alpha*drnorm) - erfc(etaij*drnorm))
                    pot_ij =  erfdiff/drnorm
                    vi = vi + Q_i*Q_j*pot_ij
                 end if
              end do
           end do
        end do
     end if
  end do
  h = vi
end subroutine energy_pim_@PBC@
